#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
CASSCF analytical nuclear gradients

Ref.
J. Comput. Chem., 5, 589

MRH: copied from pyscf.grad.casscf.py on 12/07/2019
Contains my modifications for SA-CASSCF gradients
1. Generalized Fock has nonzero i->a and u->a
2. Memory footprint for differentiated eris bugfix
'''

import inspect
import time
from functools import reduce
from itertools import product
import numpy
from scipy import linalg
from pyscf import gto
from pyscf import lib
from pyscf import ao2mo
from pyscf.lib import logger
from pyscf.grad import casci as casci_grad
from pyscf.grad import rhf as rhf_grad
from pyscf.grad.mp2 import _shell_prange
from pyscf.df.grad import rhf as dfrhf_grad
from pyscf.df.grad.casdm2_util import (solve_df_rdm2, grad_elec_dferi,
                                       grad_elec_auxresponse_dferi)
from pyscf.mcscf.addons import StateAverageMCSCFSolver

def grad_elec(mc_grad, mo_coeff=None, ci=None, atmlst=None, verbose=None):
    mc = mc_grad.base
    with_df = mc.with_df
    if mo_coeff is None: mo_coeff = mc.mo_coeff
    if ci is None: ci = mc.ci
    if mc.frozen is not None:
        raise NotImplementedError

    time0 = logger.process_clock(), logger.perf_counter()
    log = logger.new_logger(mc_grad, verbose)
    mol = mc_grad.mol
    ncore = mc.ncore
    ncas = mc.ncas
    nocc = ncore + ncas
    nelecas = mc.nelecas
    nao, nmo = mo_coeff.shape

    # Necessary kludge because gfock isn't zero in occ-virt space in SA-CASSCf
    # Among many other potential applications!
    if hasattr (mc, '_tag_gfock_ov_nonzero'):
        if mc._tag_gfock_ov_nonzero:
            nocc = nmo

    mo_occ = mo_coeff[:,:nocc]
    mo_core = mo_coeff[:,:ncore]
    mo_cas = mo_coeff[:,ncore:ncore+ncas]

    casdm1, casdm2 = mc.fcisolver.make_rdm12(ci, ncas, nelecas)

# gfock = Generalized Fock, Adv. Chem. Phys., 69, 63
    dm_core = numpy.dot(mo_core, mo_core.T) * 2
    dm_cas = reduce(numpy.dot, (mo_cas, casdm1, mo_cas.T))
    # MRH flag: this is one of my kludges
    # It would be better to just pass the ERIS object used in orbital optimization
    # But I am too lazy at the moment
    aapa = with_df.ao2mo ((mo_cas, mo_cas, mo_occ, mo_cas), compact=False)
    aapa = aapa.reshape(ncas,ncas,nocc,ncas)
    vj, vk = mc._scf.get_jk(mol, (dm_core, dm_cas))
    h1 = mc.get_hcore()
    vhf_c = vj[0] - vk[0] * .5
    vhf_a = vj[1] - vk[1] * .5
    gfock = numpy.zeros ((nocc, nocc))
    gfock[:,:ncore] = reduce(numpy.dot, (mo_occ.T, h1 + vhf_c + vhf_a, mo_core)) * 2
    gfock[:,ncore:ncore+ncas] = reduce(numpy.dot, (mo_occ.T, h1 + vhf_c, mo_cas, casdm1))
    gfock[:,ncore:ncore+ncas] += numpy.einsum('uviw,vuwt->it', aapa, casdm2)
    dme0 = reduce(numpy.dot, (mo_occ, (gfock+gfock.T)*.5, mo_occ.T))
    aapa = vj = vk = vhf_c = vhf_a = h1 = gfock = None

    dm1 = dm_core + dm_cas
    vj, vk = mc_grad.get_jk(mol, (dm_core, dm_cas))
    vhf1c, vhf1a = vj - vk * .5
    hcore_deriv = mc_grad.hcore_generator(mol)
    s1 = mc_grad.get_ovlp(mol)

    dfcasdm2 = casdm2 = solve_df_rdm2 (mc_grad, mo_cas=mo_cas, casdm2=casdm2)
    if atmlst is None:
        atmlst = range(mol.natm)
    aoslices = mol.aoslice_by_atom()
    de = grad_elec_dferi (mc_grad, mo_cas=mo_cas, dfcasdm2=dfcasdm2, atmlst=atmlst,
        max_memory=mc_grad.max_memory)[0]
    if mc_grad.auxbasis_response:
        de_aux = vj.aux - vk.aux * .5
        de_aux = de_aux.sum ((0,1)) - de_aux[1,1]
        de_aux += grad_elec_auxresponse_dferi (mc_grad, mo_cas=mo_cas, dfcasdm2=dfcasdm2,
            atmlst=atmlst, max_memory=mc_grad.max_memory)[0]
        de += de_aux
    dfcasdm2 = casdm2 = None

    for k, ia in enumerate(atmlst):
        shl0, shl1, p0, p1 = aoslices[ia]
        h1ao = hcore_deriv(ia)
        de[k] += numpy.einsum('xij,ij->x', h1ao, dm1)
        de[k] -= numpy.einsum('xij,ij->x', s1[:,p0:p1], dme0[p0:p1]) * 2
        de[k] += numpy.einsum('xij,ij->x', vhf1c[:,p0:p1], dm1[p0:p1]) * 2
        de[k] += numpy.einsum('xij,ij->x', vhf1a[:,p0:p1], dm_core[p0:p1]) * 2

    log.timer('CASSCF nuclear gradients', *time0)
    return de

def as_scanner(mcscf_grad):
    '''Generating a nuclear gradients scanner/solver (for geometry optimizer).

    The returned solver is a function. This function requires one argument
    "mol" as input and returns energy and first order nuclear derivatives.

    The solver will automatically use the results of last calculation as the
    initial guess of the new calculation.  All parameters assigned in the
    nuc-grad object and SCF object (DIIS, conv_tol, max_memory etc) are
    automatically applied in the solver.

    Note scanner has side effects.  It may change many underlying objects
    (_scf, with_df, with_x2c, ...) during calculation.

    Examples:

    >>> from pyscf import gto, scf, mcscf
    >>> mol = gto.M(atom='N 0 0 0; N 0 0 1.1', verbose=0)
    >>> mc_grad_scanner = mcscf.CASSCF(scf.RHF(mol), 4, 4).nuc_grad_method().as_scanner()
    >>> etot, grad = mc_grad_scanner(gto.M(atom='N 0 0 0; N 0 0 1.1'))
    >>> etot, grad = mc_grad_scanner(gto.M(atom='N 0 0 0; N 0 0 1.5'))
    '''
    if isinstance(mcscf_grad, lib.GradScanner):
        return mcscf_grad

    logger.info(mcscf_grad, 'Create scanner for %s', mcscf_grad.__class__)
    name = mcscf_grad.__class__.__name__ + CASSCF_GradScanner.__name_mixin__
    return lib.set_class(CASSCF_GradScanner(mcscf_grad),
                         (CASSCF_GradScanner, mcscf_grad.__class__), name)

class CASSCF_GradScanner(lib.GradScanner):
    def __init__(self, g):
        lib.GradScanner.__init__(self, g)

    def __call__(self, mol_or_geom, **kwargs):
        if isinstance(mol_or_geom, gto.MoleBase):
            assert mol_or_geom.__class__ == gto.Mole
            mol = mol_or_geom
        else:
            mol = self.mol.set_geom_(mol_or_geom, inplace=False)

        mc_scanner = self.base
        e_tot = mc_scanner(mol)
        if isinstance(mc_scanner, StateAverageMCSCFSolver):
            e_tot = mc_scanner.e_average

        self.mol = mol
        de = self.kernel(**kwargs)
        return e_tot, de


class Gradients(casci_grad.Gradients):
    '''Non-relativistic restricted Hartree-Fock gradients'''

    _keys = {'with_df', 'auxbasis_response'}

    def __init__(self, mc):
        self.with_df = mc.with_df
        self.auxbasis_response = True
        casci_grad.Gradients.__init__(self, mc)

    grad_elec = grad_elec

    def get_jk (self, mol=None, dm=None, hermi=0):
        if mol is None: mol = self.mol
        if dm is None: dm = self.base.make_rdm1()
        cpu0 = (logger.process_clock(), logger.perf_counter())
        vj, vk = dfrhf_grad.get_jk(self, mol, dm)
        logger.timer(self, 'vj and vk', *cpu0)
        return vj, vk

    def kernel (self, mo_coeff=None, ci=None, atmlst=None, verbose=None):
        log = logger.new_logger(self, verbose)
        if atmlst is None:
            atmlst = self.atmlst
        else:
            self.atmlst = atmlst

        if self.verbose >= logger.WARN:
            self.check_sanity()
        if self.verbose >= logger.INFO:
            self.dump_flags()

        de = self.grad_elec(mo_coeff, ci, atmlst, log)
        self.de = de = de + self.grad_nuc(atmlst=atmlst)
        if self.mol.symmetry:
            self.de = self.symmetrize(self.de, atmlst)
        self._finalize()
        return self.de

    def _finalize(self):
        if self.verbose >= logger.NOTE:
            logger.note(self, '--------------- %s gradients ---------------',
                        self.base.__class__.__name__)
            self._write(self.mol, self.de, self.atmlst)
            logger.note(self, '----------------------------------------------')

    as_scanner = as_scanner

    to_gpu = lib.to_gpu

Grad = Gradients



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np
from scipy import linalg
from pyscf import ao2mo, lib
from pyscf.lib import logger
from pyscf.mcscf import newton_casscf
import copy
from functools import reduce
from pyscf.mcpdft.cmspdft import coulomb_tensor

# TODO: docstring?
def diab_response (mc_grad, Lis, mo=None, ci=None, eris=None, **kwargs):
    '''Computes the Hessian-vector product of

    Q_a-a = 1/2 sum_I g_pqrs <I|p'q|I> <I|r's|I>

    where the vector is a vector of intermediate-state rotations and the
    external derivatives are with respect to orbital rotations and CI
    transfers.

    Args:
        mc_grad : object of class Gradients (CASSCF or CASCI)
        Lis : ndarray of shape (nroots*(nroots-1)/2,)
            Contains step vector for intermediate state rotations

    Kwargs:
        mo : ndarray of shape (nao,nmo)
            Contains MO coefficients
        ci : ndarray or list of length (nroots)
            Contains intermediate-state CI vectors
        eris : object of class ERIS (CASSCF or CASCI)
            Contains (true) ERIs in the MO basis

    Returns:
        R : ndarray of shape (mc_grad.ngorb+mc_grad.nci)
            Contains Hessian-vector product
    '''

    mc = mc_grad.base
    if mo is None: mo = mc.mo_coeff
    if ci is None: ci = mc.ci
    if eris is None: eris = mc.ao2mo (mo)
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    nroots, nocc = mc_grad.nroots, ncore + ncas
    nmo = mo.shape[1]

    # CI vector shift
    L = np.zeros ((nroots, nroots), dtype=Lis.dtype)
    L[np.tril_indices (nroots, k=-1)] = Lis[:]
    L -= L.T
    ci_arr = np.asarray (ci)
    Lci = np.tensordot (L, ci_arr, axes=1)

    # Density matrices
    tril_idx = np.tril_indices (nroots)
    diag_idx = np.arange (nroots)
    diag_idx = diag_idx * (diag_idx+1) // 2 + diag_idx
    tdm1 = np.stack (mc.fcisolver.states_trans_rdm12 (ci_arr[tril_idx[0]],
        ci_arr[tril_idx[1]], ncas, nelecas)[0], axis=0)
    dm1 = tdm1[diag_idx,:,:]
    edm1 = np.stack (mc.fcisolver.states_trans_rdm12 (Lci, ci, ncas,
        nelecas)[0], axis=0)
    edm1 += edm1.transpose (0,2,1)

    # Potentials
    aapa = np.zeros ([ncas,ncas,nmo,ncas], dtype=dm1.dtype)
    for i in range (ncas):
        j = i + ncore
        aapa[i,:,:,:] = eris.papa[j][:,:,:]
    vj = np.tensordot (dm1, aapa, axes=2)
    evj = np.tensordot (edm1, aapa, axes=2)

    # Orbital degree of freedom
    Rorb = np.zeros ((nmo,nmo), dtype=vj[0].dtype)
    Rorb[:,ncore:nocc] = sum ([np.dot (v, ed) + np.dot (ev, d)
        for v, d, ev, ed in zip (vj, dm1, evj, edm1)])
    Rorb -= Rorb.T

    # CI degree of freedom
    w = coulomb_tensor (mc, mo_coeff=mo, ci=ci, h2eff=aapa[:,:,ncore:nocc,:])
    const_IJ = -4*np.einsum ('jiik,ik->ij', w, L)
    const_IJ -= 2*np.einsum ('iijk,ik->ij', w, L)
    const_IJ += 2*np.einsum ('jkkk,ik->ij', w, L)
    Rci = np.tensordot (const_IJ, ci_arr, axes=1) # Delta_IJ |J> term

    def contract (v,c):
        return mc.fcisolver.contract_1e (v, c, ncas, nelecas)

    vj, evj = vj[:,ncore:nocc,:], evj[:,ncore:nocc,:]
    vci = np.stack ([contract (v,c) for v, c in zip (vj, ci)], axis=0)
    Rci -= 2 * np.tensordot (L, vci, axes=1) # -2 |zW_I> term
    for I in range (nroots):
        Rci[I] += 2 * contract (vj[I], Lci[I]) # 2 W^I_I |z_I> term
        Rci[I] += 2 * contract (evj[I], ci[I]) # 4 W^zI_I |I> z_IJ term
        # (*2 in def. of evj)

    return mc_grad.pack_uniq_var (2*Rorb, Rci)

# TODO: get rid?? Fix?? Unittest???
# BROKEN FOR CI AND IS; DO NOT USE
def diab_response_o0 (mc_grad, Lis, mo=None, ci=None, eris=None, **kwargs):
    '''Alternate implementation: monkeypatch everything but
    active-active Coulomb part of the Hamiltonian and call
    newton_casscf.gen_g_hop ()[2].
    '''

    mc = mc_grad.base
    if mo is None: mo = mc.mo_coeff
    if ci is None: ci = mc.ci
    if eris is None: eris = mc.ao2mo (mo)
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    nroots, nocc, nmo = mc_grad.nroots, ncore + ncas, mo.shape[1]
    moH = mo.conj ().T

    # CI vector shift
    L = np.zeros ((nroots, nroots), dtype=Lis.dtype)
    L[np.tril_indices (nroots, k=-1)] = Lis[:]
    L -= L.T
    ci_arr = np.asarray (ci)
    Lci = list (np.tensordot (L, ci_arr, axes=1))
    x = mc_grad.pack_uniq_var (np.zeros ((nmo,nmo)), Lci)

    # Fake Hamiltonian!
    h1e_mo = moH @ mc.get_hcore () @ mo
    feris = mc.ao2mo (mo)
    for i in range (nmo):
        feris.papa[i][:,:,:] = 0.0
        feris.ppaa[i][:ncore,:,:] = 0.0
        feris.ppaa[i][nocc:,:,:] = 0.0
    feris.vhf_c[:,:] = -h1e_mo.copy ()
    from pyscf.mcscf.newton_casscf import _pack_ci_get_H as getH
    from pyscf.mcscf import addons
    def _pack_ci_get_H (mc1, mo1, ci1):
        ci1, _, _Hdiag, linkstrl, linkstr, _pack_ci, _unpack_ci = getH (mc1,
            mo1, ci1)
        dm1 = mc.fcisolver.states_make_rdm1 (ci1, ncas, nelecas)
        _state_arg = addons.StateAverageMixFCISolver_state_args
        def _Hci_kernel (s, op1, op2, ci1, ci2, ne, my_L):
            h1ci2, h1ci1, h2ci1, c1c2 = [], [], [], []
            for o1, o2, c2, c1 in zip (op1, op2, ci2, ci1):
                h1ci2.append (s.contract_1e (o1, c2, ncas, ne))
                h2ci1.append (s.contract_1e (o2, c1, ncas, ne))
                h1ci1.append (s.contract_1e (o1, c1, ncas, ne))
                c1c2.append (c1.dot (c2))
            if np.all (np.asarray (c1c2) < 0.5): # chain rule
                h1ci2, h2ci1 = np.asarray (h1ci2), np.asarray (h2ci1)
                h1ci1 = np.tensordot (my_L, np.asarray (h1ci1), axes=1)
                return list (h1ci2 + h2ci1 - h1ci1)
            else:
                return h1ci2

        if isinstance (mc.fcisolver, addons.StateAverageMixFCISolver):
            full_idx = np.arange (nroots)
            def _Hci (h1, h2, ci2):
                hci = []
                tm1 = mc.fcisolver.states_trans_rdm12 (ci2, ci1, ncas,
                    nelecas)[9]
                for s, args, kwargs in enumerate (mc.fcisolver._loop_solver (
                        _state_arg (ci2), _state_arg (ci1), _state_arg (dm1),
                        _state_arg (tm1), _state_arg (full_idx))):
                    ci2i, ci1i, dm1i, tm1i, idx = args[0:5]
                    Lsec = L[np.ix_(idx,idx)]
                    nelec = mc.fcisolver._get_nelec (s, nelecas)
                    op1 = h1[None,:,:] + np.tensordot (dm1i, h2, axes=2)
                    op2 = np.tensordot (tm1i + tm1i.transpose (0,2,1), h2,
                        axes=2)
                    hci.extend (_Hci_kernel (s, op1, op2, ci1i, ci2i, nelec,
                        Lsec))
                return hci
        else:
            def _Hci (h1, h2, ci2):
                tm1 = np.asarray (mc.fcisolver.states_trans_rdm12 (ci2, ci1,
                    ncas, nelecas)[0])
                op1 = h1[None,:,:] + np.tensordot (dm1, h2, axes=2)
                op2 = np.tensordot (tm1 + tm1.transpose (0,2,1), h2, axes=2)
                return _Hci_kernel (mc.fcisolver, op1, op2, ci1, ci2, nelecas,
                    L)
        return ci1, _Hci, _Hdiag, linkstrl, linkstr, _pack_ci, _unpack_ci

    # Fake 2TDM!
    dm1 = mc.fcisolver.states_make_rdm1 (ci, ncas, nelecas)
    def trans_rdm12 (ci1, ci0, *args, **kwargs):
        tm1, tm2 = mc.fcisolver.states_trans_rdm12 (ci1, ci0, *args, **kwargs)
        for t1, t2, d1, w in zip (tm1, tm2, dm1, mc.weights):
            t2[:,:,:,:] = w * (np.multiply.outer (t1, d1)
                             + np.multiply.outer (d1, t1))
            t1[:,:] *= w
        return sum (tm1), sum (tm2)

    # Fake Newton CASSCF!
    with lib.temporary_env (newton_casscf, _pack_ci_get_H=_pack_ci_get_H):
        with lib.temporary_env (mc.fcisolver, trans_rdm12=trans_rdm12):
            hx = newton_casscf.gen_g_hop (mc, mo, ci, feris, verbose=0)[2](x)
    hx *= nroots
    hx_orb, hx_ci = mc_grad.unpack_uniq_var (hx)
    hx_ci = np.asarray (hx_ci)
    hx_is = lib.einsum ('pab,qab->pq', hx_ci, ci_arr.conj ())
    hx_ci -= np.tensordot(hx_is, ci_arr, axes=1)

    return mc_grad.pack_uniq_var (hx_orb, hx_ci)

def diab_grad (mc_grad, Lis, atmlst=None, mo=None, ci=None, eris=None,
        mf_grad=None, **kwargs):
    '''Computes the partial first derivatives of

    Q_a-a = 1/2 sum_I g_pqrs <I|p'q|I> <I|r's|I>

    with respect to geometry perturbation.

    Args:
        mc_grad : object of class Gradients (CASSCF or CASCI)
        Lis : ndarray of shape (nroots*(nroots-1)/2,)
            Contains step vector for intermediate state rotations

    Kwargs:
        atmlst : list
            List of atoms whose geometries are perturbed. Defaults
            to all atoms in mc_grad.mol.
        mo : ndarray of shape (nao,nmo)
            Contains MO coefficients
        ci : ndarray or list of length (nroots)
            Contains intermediate-state CI vectors
        eris : object of class ERIS (CASSCF or CASCI)
            Contains (true) ERIs in the MO basis
        mf_grad: object of class Gradients (RHF)
            Defaults to mc_grad.base.get_rhf_base ().nuc_grad_method ()

    Returns:
        de : ndarray of shape (len (atmlst), 3)
            Contains gradient vector
    '''

    mc = mc_grad.base
    mol = mc_grad.mol
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    nroots, nocc, nmo = mc_grad.nroots, ncore + ncas, mo.shape[1]
    moH = mo.conj ().T
    mo_cas = mo[:,ncore:nocc]
    moH_cas = moH[ncore:nocc,:]
    if mf_grad is None: mf_grad = mc.get_rhf_base ().nuc_grad_method()
    if atmlst is None: atmlst = list (range(mol.natm))

    # CI vector shift
    L = np.zeros ((nroots, nroots), dtype=Lis.dtype)
    L[np.tril_indices (nroots, k=-1)] = Lis[:]
    L -= L.T
    ci_arr = np.asarray (ci)
    Lci = np.tensordot (L, ci_arr, axes=1)

    # Density matrices
    dm1 = np.stack (mc.fcisolver.states_make_rdm1 (ci, ncas, nelecas), axis=0)
    edm1 = np.stack (mc.fcisolver.states_trans_rdm12 (Lci, ci, ncas,
        nelecas)[0], axis=0)
    edm1 += edm1.transpose (0,2,1)
    dm1_ao = reduce (np.dot, (mo_cas, dm1, moH_cas)).transpose (1,0,2)
    edm1_ao = reduce (np.dot, (mo_cas, edm1, moH_cas)).transpose (1,0,2)

    # Potentials and operators
    aapa = np.zeros ([nmo,]+[ncas,]*3, dtype=dm1.dtype)
    for i in range (nmo): aapa[i] = eris.ppaa[i][ncore:nocc,:,:]
    aapa = aapa.transpose (2,3,0,1)
    vj = np.tensordot (dm1, aapa, axes=2)
    evj = np.tensordot (edm1, aapa, axes=2)
    dvj_all = mf_grad.get_j (mc.mol, list(dm1_ao) + list(edm1_ao))
    dvj_aux = getattr (dvj_all, 'aux', np.zeros ((nroots, nroots, mol.natm,
        3)))
    dvj = np.stack (dvj_all[:nroots], axis=1)
    devj = np.stack (dvj_all[nroots:], axis=1)

    # Generalized Fock and overlap operator
    gfock = np.zeros ([nmo,nmo], dtype=vj.dtype)
    gfock[:,ncore:nocc] = sum ([np.dot (v, ed) + np.dot (ev, d)
        for v, d, ev, ed in zip (vj, dm1, evj, edm1)])
    dme0 = reduce (np.dot, (mo, (gfock+gfock.T)*.5, moH))
    s1 = mf_grad.get_ovlp (mc.mol)

    # Crunch
    de_direct = np.zeros ((len (atmlst), 3))
    de_renorm = np.zeros ((len (atmlst), 3))
    aoslices = mol.aoslice_by_atom()
    for k, ia in enumerate(atmlst):
        shl0, shl1, p0, p1 = aoslices[ia]
        de_renorm[k] -= lib.einsum('xpq,pq->x', s1[:,p0:p1], dme0[p0:p1]) * 2
        de_direct[k] += lib.einsum('xipq,ipq->x', dvj[:,:,p0:p1],
            edm1_ao[:,p0:p1]) * 2
        de_direct[k] += lib.einsum('xipq,ipq->x', devj[:,:,p0:p1],
            dm1_ao[:,p0:p1]) * 2
    dvj_aux = dvj_aux[:,:,atmlst,:]
    de_aux = (np.trace (dvj_aux, offset=nroots, axis1=0, axis2=1)
            + np.trace (dvj_aux, offset=-nroots, axis1=0, axis2=1))

    logger.debug (mc, "CMS-PDFT Lis lagrange direct component:\n{}".format (
        de_direct))
    logger.debug (mc, "CMS-PDFT Lis lagrange renorm component:\n{}".format (
        de_renorm))
    logger.debug (mc, "CMS-PDFT Lis lagrange auxbasis component:\n{}".format (
        de_aux))
    de = de_direct + de_aux + de_renorm
    return de

# TODO: get rid? Unittest?
def diab_grad_o0 (mc_grad, Lis, atmlst=None, mo=None, ci=None, eris=None,
        mf_grad=None, **kwargs):
    ''' Monkeypatch version of diab_grad '''
    mc = mc_grad.base
    ncas, nelecas, nroots = mc.ncas, mc.nelecas, mc_grad.nroots
    if mf_grad is None: mf_grad = mc.get_rhf_base ().nuc_grad_method()

    # CI vector shift
    L = np.zeros ((nroots, nroots), dtype=Lis.dtype)
    L[np.tril_indices (nroots, k=-1)] = Lis[:]
    L -= L.T
    ci_arr = np.asarray (ci)
    Lci = list (np.tensordot (L, ci_arr, axes=1))

    # Fake dms!
    dm1 = mc.fcisolver.states_make_rdm1 (ci, ncas, nelecas)
    def trans_rdm12 (ci1, ci0, *args, **kwargs):
        tm1, tm2 = mc.fcisolver.states_trans_rdm12 (ci1, ci0, *args, **kwargs)
        for t1, t2, d1, w in zip (tm1, tm2, dm1, mc.weights):
            t2[:,:,:,:] = (np.multiply.outer (t1, d1)
                         + np.multiply.outer (d1, t1))
            t1[:,:] = 0.0
        return sum (tm1), sum (tm2)

    from pyscf.grad.sacasscf import Lci_dot_dgci_dx
    with lib.temporary_env (mc.fcisolver, trans_rdm12=trans_rdm12):
        de = Lci_dot_dgci_dx (Lci, mc.weights, mc, mo_coeff=mo, ci=ci,
            atmlst=atmlst, eris=eris, mf_grad=mf_grad)
    return de

if __name__ == '__main__':
    import math
    from pyscf import scf, gto, mcscf
    from pyscf.fci import csf_solver
    from pyscf.mcpdft.pdft_feff import vector_error
    xyz = '''O  0.00000000   0.08111156   0.00000000
             H  0.78620605   0.66349738   0.00000000
             H -0.78620605   0.66349738   0.00000000'''
    mol = gto.M (atom=xyz, basis='6-31g', symmetry=False, output='cmspdft.log',
        verbose=lib.logger.DEBUG)
    mf = scf.RHF (mol).run ()
    mc = mcscf.CASSCF (mf, 4, 4).set (fcisolver = csf_solver (mol, 1))
    mc = mc.state_average ([1.0/3,]*3).run ()
    ci_arr = np.asarray (mc.ci)

    mc_grad = mc.nuc_grad_method ()
    Lis = math.pi * (np.random.rand (3) - 0.5)
    eris = mc.ao2mo (mc.mo_coeff)

    dw_test = diab_response (mc_grad, Lis, mo=mc.mo_coeff, ci=mc.ci,
        eris=eris)
    dworb_test, dwci_test = mc_grad.unpack_uniq_var (dw_test)
    dwci_test = np.asarray (dwci_test)
    dwis_test = lib.einsum ('pab,qab->pq', dwci_test, ci_arr.conj ())
    dwci_test -= lib.einsum ('pq,qab->pab', dwis_test, ci_arr)
    dw_ref = diab_response_o0 (mc_grad, Lis, mo=mc.mo_coeff, ci=mc.ci,
        eris=eris)
    dworb_ref, dwci_ref = mc_grad.unpack_uniq_var (dw_ref)
    dwci_ref = np.asarray (dwci_ref)
    dwis_ref = lib.einsum ('pab,qab->pq', dwci_ref, ci_arr.conj ())
    dwci_ref -= lib.einsum ('pq,qab->pab', dwis_ref, ci_arr)
    dh_test = diab_grad (mc_grad, Lis, mo=mc.mo_coeff, ci=mc.ci, eris=eris)
    dh_ref = diab_grad_o0 (mc_grad, Lis, mo=mc.mo_coeff, ci=mc.ci, eris=eris)

    print ("dworb:", vector_error (dworb_test, dworb_ref), linalg.norm (
        dworb_ref))
    print ("dwci:", vector_error (dwci_test, dwci_ref), linalg.norm (dwci_ref))
    print ("dwis:", vector_error (dwis_test, dwis_ref), linalg.norm (dwis_ref))
    print ("dh:", vector_error (dh_test, dh_ref), linalg.norm (dh_ref))

    print (dh_test, '\n', dh_ref)



#!/usr/bin/env python
# Copyright 2014-2018,2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Analytical Fourier transformation AO-pair product for PBC
'''

import ctypes
import numpy as np
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.pbc import gto as pbcgto
from pyscf.gto.ft_ao import ft_ao as mol_ft_ao
from pyscf.pbc.tools import k2gamma
from pyscf.pbc.tools import pbc as pbctools
from pyscf.pbc.lib.kpts_helper import is_zero, gamma_point
from pyscf import __config__

RCUT_THRESHOLD = getattr(__config__, 'pbc_scf_rsjk_rcut_threshold', 1.0)
# kecut=10 can roughly converge GTO with alpha=0.5
KECUT_THRESHOLD = getattr(__config__, 'pbc_scf_rsjk_kecut_threshold', 10.0)

STEEP_BASIS = 0
LOCAL_BASIS = 1
SMOOTH_BASIS = 2

libpbc = lib.load_library('libpbc')

#
# \int mu*nu*exp(-ik*r) dr
#
def ft_aopair(cell, Gv, shls_slice=None, aosym='s1',
              b=None, gxyz=None, Gvbase=None, kpti_kptj=np.zeros((2,3)),
              q=None, intor='GTO_ft_ovlp', comp=1, verbose=None):
    r'''
    Fourier transform AO pair for a pair of k-points
    \sum_T exp(-i k_j * T) \int exp(-i(G+q)r) i(r) j(r-T) dr^3
    '''
    kpti, kptj = kpti_kptj
    if q is None:
        q = kptj - kpti
    val = ft_aopair_kpts(cell, Gv, shls_slice, aosym, b, gxyz, Gvbase,
                         q, kptj.reshape(1,3), intor, comp)
    return val[0]


def ft_aopair_kpts(cell, Gv, shls_slice=None, aosym='s1',
                   b=None, gxyz=None, Gvbase=None, q=np.zeros(3),
                   kptjs=np.zeros((1,3)), intor='GTO_ft_ovlp', comp=1,
                   bvk_kmesh=None, out=None):
    r'''
    Fourier transform AO pair for a group of k-points
    \sum_T exp(-i k_j * T) \int exp(-i(G+q)r) i(r) j(r-T) dr^3

    The return array holds the AO pair
    corresponding to the kpoints given by kptjs
    '''
    log = logger.new_logger(cell)
    kptjs = np.asarray(kptjs, order='C').reshape(-1,3)

    rs_cell = _RangeSeparatedCell.from_cell(cell, KECUT_THRESHOLD,
                                            RCUT_THRESHOLD, log)
    if bvk_kmesh is None:
        bvk_kmesh = k2gamma.kpts_to_kmesh(cell, kptjs)
        log.debug2('Set bvk_kmesh = %s', bvk_kmesh)
    rcut = estimate_rcut(rs_cell)
    supmol = ExtendedMole.from_cell(rs_cell, bvk_kmesh, rcut.max(), log)
    supmol = supmol.strip_basis(rcut)

    ft_kern = supmol.gen_ft_kernel(aosym, intor=intor, comp=comp,
                                   return_complex=True, verbose=log)

    return ft_kern(Gv, gxyz, Gvbase, q, kptjs, shls_slice)


@lib.with_doc(mol_ft_ao.__doc__)
def ft_ao(mol, Gv, shls_slice=None, b=None,
          gxyz=None, Gvbase=None, kpt=np.zeros(3), verbose=None):
    if gamma_point(kpt):
        return mol_ft_ao(mol, Gv, shls_slice, b, gxyz, Gvbase, verbose)
    else:
        kG = Gv + kpt
        return mol_ft_ao(mol, kG, shls_slice, None, None, None, verbose)


def gen_ft_kernel(supmol, aosym='s1', intor='GTO_ft_ovlp', comp=1,
                  return_complex=False, kpts=None, verbose=None):
    r'''
    Generate the analytical fourier transform kernel for AO products

    \sum_T exp(-i k_j * T) \int exp(-i(G+q)r) i(r) j(r-T) dr^3
    '''
    log = logger.new_logger(supmol)
    cput0 = logger.process_clock(), logger.perf_counter()
    rs_cell = supmol.rs_cell
    assert isinstance(rs_cell, _RangeSeparatedCell)

    # The number of basis in the original cell
    nbasp = rs_cell.ref_cell.nbas
    cell0_ao_loc = rs_cell.ref_cell.ao_loc
    bvk_ncells, rs_nbas, nimgs = supmol.bas_mask.shape

    ovlp_mask = supmol.get_ovlp_mask()
    bvk_ovlp_mask = lib.condense('np.any', ovlp_mask, rs_cell.sh_loc, supmol.sh_loc)
    cell0_ovlp_mask = bvk_ovlp_mask.reshape(nbasp, bvk_ncells, nbasp).any(axis=1)
    ovlp_mask = ovlp_mask.astype(np.int8)
    cell0_ovlp_mask = cell0_ovlp_mask.astype(np.int8)

    if kpts is not None:
        expLk = np.exp(1j*np.dot(supmol.bvkmesh_Ls, kpts.T))
        expLkR = np.asarray(expLk.real, order='C')
        expLkI = np.asarray(expLk.imag, order='C')
        _expLk = (expLkR, expLkI)
    else:
        _expLk = None

    b = rs_cell.reciprocal_vectors()
    if abs(b-np.diag(b.diagonal())).sum() < 1e-8:
        _eval_gz = 'GTO_Gv_orth'
    else:
        _eval_gz = 'GTO_Gv_nonorth'
    drv = libpbc.PBC_ft_bvk_drv
    cintor = getattr(libpbc, rs_cell._add_suffix(intor))

    log.timer_debug1('ft_ao kernel initialization', *cput0)

    # TODO: use Gv = b * gxyz + q in c code
    # TODO: add zfill
    def ft_kernel(Gv, gxyz=None, Gvbase=None, q=np.zeros(3), kptjs=None,
                  shls_slice=None, aosym=aosym, out=None):
        '''
        Analytical FT for orbital products. The output tensor has the shape [nGv, nao, nao]
        '''
        cput0 = logger.process_clock(), logger.perf_counter()
        assert q.ndim == 1
        if kptjs is None:
            if _expLk is None:
                expLkR = np.ones((nimgs,1))
                expLkI = np.zeros((nimgs,1))
            else:
                expLkR, expLkI = _expLk
        else:
            kptjs = np.asarray(kptjs, order='C').reshape(-1,3)
            expLk = np.exp(1j*np.dot(supmol.bvkmesh_Ls, kptjs.T))
            expLkR = np.asarray(expLk.real, order='C')
            expLkI = np.asarray(expLk.imag, order='C')
            expLk = None

        nkpts = expLkR.shape[1]
        GvT = np.asarray(Gv.T + q[:,None], order='C')
        nGv = GvT.shape[1]

        if shls_slice is None:
            shls_slice = (0, nbasp, 0, nbasp)
        ni = cell0_ao_loc[shls_slice[1]] - cell0_ao_loc[shls_slice[0]]
        nj = cell0_ao_loc[shls_slice[3]] - cell0_ao_loc[shls_slice[2]]
        shape = (nkpts, comp, ni, nj, nGv)

        aosym = aosym[:2]
        if aosym == 's1hermi':
            # Gamma point only
            assert is_zero(q) and is_zero(kptjs) and ni == nj
            # Theoretically, hermitian symmetry can be also found for kpti == kptj != 0:
            #       f_ji(G) = \int f_ji exp(-iGr) = \int f_ij^* exp(-iGr) = [f_ij(-G)]^*
            # hermi operation needs to reorder axis-0.  It is inefficient.
        elif aosym == 's2':
            i0 = cell0_ao_loc[shls_slice[0]]
            i1 = cell0_ao_loc[shls_slice[1]]
            nij = i1*(i1+1)//2 - i0*(i0+1)//2
            shape = (nkpts, comp, nij, nGv)

        if gxyz is None or Gvbase is None or (abs(q).sum() > 1e-9):
            p_gxyzT = lib.c_null_ptr()
            p_mesh = (ctypes.c_int*3)(0,0,0)
            p_b = (ctypes.c_double*1)(0)
            eval_gz = 'GTO_Gv_general'
        else:
            eval_gz = _eval_gz
            gxyzT = np.asarray(gxyz.T, order='C', dtype=np.int32)
            p_gxyzT = gxyzT.ctypes.data_as(ctypes.c_void_p)
            bqGv = np.hstack((b.ravel(), q) + Gvbase)
            p_b = bqGv.ctypes.data_as(ctypes.c_void_p)
            p_mesh = (ctypes.c_int*3)(*[len(x) for x in Gvbase])

        eval_gz = getattr(libpbc, eval_gz)
        if nkpts == 1:
            fill = getattr(libpbc, 'PBC_ft_bvk_nk1'+aosym)
        else:
            fill = getattr(libpbc, 'PBC_ft_bvk_k'+aosym)

        if return_complex:
            fsort = getattr(libpbc, 'PBC_ft_zsort_' + aosym)
            out = np.ndarray(shape, dtype=np.complex128, buffer=out)
        else:
            fsort = getattr(libpbc, 'PBC_ft_dsort_' + aosym)
            out = np.ndarray((2,) + shape, buffer=out)

        if nGv > 0:
            drv(cintor, eval_gz, fill, fsort,
                out.ctypes.data_as(ctypes.c_void_p),
                expLkR.ctypes.data_as(ctypes.c_void_p),
                expLkI.ctypes.data_as(ctypes.c_void_p),
                ctypes.c_int(bvk_ncells), ctypes.c_int(nimgs),
                ctypes.c_int(nkpts), ctypes.c_int(nbasp), ctypes.c_int(comp),
                supmol.seg_loc.ctypes.data_as(ctypes.c_void_p),
                supmol.seg2sh.ctypes.data_as(ctypes.c_void_p),
                cell0_ao_loc.ctypes.data_as(ctypes.c_void_p),
                (ctypes.c_int*4)(*shls_slice),
                ovlp_mask.ctypes.data_as(ctypes.c_void_p),
                cell0_ovlp_mask.ctypes.data_as(ctypes.c_void_p),
                GvT.ctypes.data_as(ctypes.c_void_p), p_b, p_gxyzT, p_mesh, ctypes.c_int(nGv),
                supmol._atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol.natm),
                supmol._bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(supmol.nbas),
                supmol._env.ctypes.data_as(ctypes.c_void_p))
            log.timer_debug1(f'ft_ao intor {intor}', *cput0)

        if return_complex:
            if aosym == 's1hermi':
                for i in range(1, ni):
                    out[:,:,:i,i] = out[:,:,i,:i]
            out = np.rollaxis(out, -1, 2)
            if comp == 1:
                out = out[:,0]
            return out
        else:
            if aosym == 's1hermi':
                for i in range(1, ni):
                    out[:,:,:,:i,i] = out[:,:,:,i,:i]
            out = np.rollaxis(out, -1, 3)
            if comp == 1:
                out = out[:,:,0]
            return out

    return ft_kernel


class _RangeSeparatedCell(pbcgto.Cell):
    '''Cell with partially de-contracted basis'''
    def __init__(self):
        # ref_cell is the original cell of which the basis to be de-contracted
        self.ref_cell = None
        # For each de-contracted basis, the shell Id in the original cell
        self.bas_map = None
        # Type of each de-contracted basis
        self.bas_type = None
        # Each shell in the original cell can have several segments in the rs-cell.
        # sh_loc indicates the shell Id in rs-cell for each shell in cell.
        self.sh_loc = None

    @classmethod
    def from_cell(cls, cell, ke_cut_threshold=None, rcut_threshold=None,
                  in_rsjk=False, verbose=None):
        from pyscf.pbc.df import aft
        rs_cell = cls()
        rs_cell.__dict__.update(cell.__dict__)
        rs_cell.ref_cell = cell

        if ke_cut_threshold is None:
            rs_cell.bas_map = np.arange(cell.nbas, dtype=np.int32)
            rs_cell.bas_type = np.empty(cell.nbas, dtype=np.int32)
            rs_cell.bas_type[:] = LOCAL_BASIS
            rs_cell.sh_loc = np.arange(cell.nbas + 1, dtype=np.int32)
            return rs_cell

        log = logger.new_logger(cell, verbose)
        if not isinstance(ke_cut_threshold, float):
            ke_cut_threshold = np.min(ke_cut_threshold)

        precision = cell.precision

        # preserves all environments defined in cell (e.g. omega, gauge origin)
        _env = cell._env.copy()
        decontracted_bas = []
        bas_type = []
        # For each basis of rs_cell, bas_map gives the basis in cell
        bas_map = []
        # For each basis of cell, bas_loc gives the first basis in rs_cell
        bas_loc = [0]

        def _append_to_decontracted_bas(orig_id, e_offset, nprim, btype):
            new_bas = cell._bas[orig_id].copy()
            new_bas[gto.PTR_EXP] += e_offset
            new_bas[gto.PTR_COEFF] += e_offset * new_bas[gto.NCTR_OF]
            new_bas[gto.NPRIM_OF] = nprim
            decontracted_bas.append(new_bas)
            bas_type.append(btype)
            bas_map.append(orig_id)

        for ib, orig_bas in enumerate(cell._bas):
            nprim = orig_bas[gto.NPRIM_OF]
            nctr = orig_bas[gto.NCTR_OF]
            l = orig_bas[gto.ANG_OF]
            es = cell.bas_exp(ib)
            # Sort exponents because integral screening of rsjk method relies on
            # the dscending order in exponents
            es_idx = es.argsort()[::-1]
            es = es[es_idx]
            cs = cell._libcint_ctr_coeff(ib)[es_idx]
            abs_cs = abs(cs).max(axis=1)

            # aft._estimate_ke_cutoff is accurate for 4c2e integrals
            # For other integrals such as nuclear attraction.
            # aft._estimate_ke_cutoff may put some primitive GTOs of large es
            # and small cs in the group SMOOTH_BASIS.  These GTOs requires a
            # large ke_cutoff (mesh) in _RSNucBuilder or _CCNucBuilder.
            if in_rsjk:
                ke = aft._estimate_ke_cutoff(es, l, abs_cs, precision)
            else:
                ke = pbcgto.cell._estimate_ke_cutoff(es, l, abs_cs, precision)

            smooth_mask = ke < ke_cut_threshold
            if rcut_threshold is None:
                local_mask = ~smooth_mask
                steep_mask = np.zeros_like(local_mask)
                rcut = None
            else:
                norm_ang = ((2*l+1)/(4*np.pi))**.5
                fac = 2*np.pi*abs_cs/cell.vol * norm_ang/es / precision
                rcut = cell.rcut
                rcut = (np.log(fac * rcut**(l+1) + 1.) / es)**.5
                rcut = (np.log(fac * rcut**(l+1) + 1.) / es)**.5
                steep_mask = (~smooth_mask) & (rcut < rcut_threshold)
                local_mask = (~steep_mask) & (~smooth_mask)

            pexp = orig_bas[gto.PTR_EXP]
            pcoeff = orig_bas[gto.PTR_COEFF]

            c_steep = cs[steep_mask]
            c_local = cs[local_mask]
            c_smooth = cs[smooth_mask]
            _env[pcoeff:pcoeff+nprim*nctr] = np.hstack([
                c_steep.T.ravel(),
                c_local.T.ravel(),
                c_smooth.T.ravel(),
            ])
            _env[pexp:pexp+nprim] = np.hstack([
                es[steep_mask],
                es[local_mask],
                es[smooth_mask],
            ])
            if log.verbose >= logger.DEBUG2:
                log.debug2('bas %d rcut %s  kecut %s', ib, rcut, ke)
                log.debug2('steep %s, %s', np.where(steep_mask)[0], es[steep_mask])
                log.debug2('local %s, %s', np.where(local_mask)[0], es[local_mask])
                log.debug2('smooth %s, %s', np.where(smooth_mask)[0], es[smooth_mask])

            nprim_steep = c_steep.shape[0]
            nprim_local = c_local.shape[0]
            nprim_smooth = c_smooth.shape[0]
            if nprim_steep > 0:
                _append_to_decontracted_bas(ib, 0, nprim_steep, STEEP_BASIS)

            if nprim_local > 0:
                _append_to_decontracted_bas(ib, nprim_steep, nprim_local, LOCAL_BASIS)

            if nprim_smooth > 0:
                _append_to_decontracted_bas(ib, nprim_steep+nprim_local,
                                            nprim_smooth, SMOOTH_BASIS)

            bas_loc.append(len(decontracted_bas))

        rs_cell._bas = np.asarray(decontracted_bas, dtype=np.int32, order='C')
        # rs_cell._bas might be of size (0, BAS_SLOTS)
        rs_cell._bas = rs_cell._bas.reshape(-1, gto.BAS_SLOTS)
        rs_cell._env = _env
        rs_cell.bas_map = np.asarray(bas_map, dtype=np.int32)
        rs_cell.bas_type = np.asarray(bas_type, dtype=np.int32)
        rs_cell.sh_loc = np.asarray(bas_loc, dtype=np.int32)
        rs_cell.ke_cutoff = ke_cut_threshold
        if log.verbose >= logger.DEBUG:
            bas_type = rs_cell.bas_type
            log.debug('rs_cell.nbas %d nao %d', rs_cell.nbas, rs_cell.nao)
            log.debug1('No. steep_bas in rs_cell %d', np.count_nonzero(bas_type == STEEP_BASIS))
            log.debug1('No. local_bas in rs_cell %d', np.count_nonzero(bas_type == LOCAL_BASIS))
            log.debug('No. smooth_bas in rs_cell %d', np.count_nonzero(bas_type == SMOOTH_BASIS))
            map_bas = rs_cell._reverse_bas_map(rs_cell.bas_map)
            log.debug2('bas_map from cell to rs_cell %s', map_bas)
            assert np.array_equiv(map_bas, bas_loc)
        log.debug2('%s.bas_type %s', cls, rs_cell.bas_type)
        log.debug2('%s.sh_loc %s', cls, rs_cell.sh_loc)
        return rs_cell

    @staticmethod
    def _reverse_bas_map(bas_map):
        '''Map basis between the original cell and the derived rs-cell.
        For each shell in the original cell, the first basis Id of the
        de-contracted basis in the rs-cell'''
        uniq_bas, map_bas = np.unique(bas_map, return_index=True)
        assert uniq_bas[-1] == len(uniq_bas) - 1
        return np.append(map_bas, len(bas_map)).astype(np.int32)

    def smooth_basis_cell(self):
        '''Construct a cell with only the smooth part of the AO basis'''
        cell_d = self.view(pbcgto.Cell)
        mask = self.bas_type == SMOOTH_BASIS
        cell_d._bas = self._bas[mask]
        segs = np.zeros(self.ref_cell.nbas)
        segs[self.bas_map[mask]] = 1
        cell_d.sh_loc = np.append(0, np.cumsum(segs)).astype(np.int32)
        logger.debug1(self, 'cell_d.nbas %d', cell_d.nbas)
        if cell_d.nbas == 0:
            return cell_d

        cell_d.ke_cutoff = ke_cutoff = pbcgto.estimate_ke_cutoff(cell_d)
        cell_d.mesh = cell_d.cutoff_to_mesh(ke_cutoff)
        logger.debug1(self, 'cell_d rcut %g ke_cutoff %g, mesh %s',
                      cell_d.rcut, ke_cutoff, cell_d.mesh)
        return cell_d

    def compact_basis_cell(self):
        '''Construct a cell with only the smooth part of the AO basis'''
        cell_c = self.copy(deep=False)
        mask = self.bas_type != SMOOTH_BASIS
        cell_c._bas = self._bas[mask]
        cell_c.bas_map = cell_c.bas_map[mask]
        cell_c.bas_type = cell_c.bas_type[mask]
        segs = self.sh_loc[1:] - self.sh_loc[:-1]
        segs[self.bas_map[~mask]] -= 1
        cell_c.sh_loc = np.append(0, np.cumsum(segs)).astype(np.int32)
        cell_c.rcut = pbcgto.estimate_rcut(cell_c, self.precision)
        return cell_c

    def merge_diffused_block(self, aosym='s1'):
        '''For AO pair that are evaluated in blocks with using the basis
        partitioning self.compact_basis_cell() and self.smooth_basis_cell(),
        merge the DD block into the CC, CD, DC blocks (C ~ compact basis,
        D ~ diffused basis)
        '''
        ao_loc = self.ref_cell.ao_loc
        smooth_bas_idx = self.bas_map[self.bas_type == SMOOTH_BASIS]
        smooth_ao_idx = self.get_ao_indices(smooth_bas_idx, ao_loc)
        nao = ao_loc[-1]
        naod = smooth_ao_idx.size
        drv = getattr(libpbc, f'PBCnr3c_fuse_dd_{aosym}')

        def merge(j3c, j3c_dd, shls_slice=None):
            if j3c_dd.size == 0:
                return j3c
            # The AO index in the original cell
            if shls_slice is None:
                slice_in_cell = (0, nao, 0, nao)
            else:
                slice_in_cell = ao_loc[list(shls_slice[:4])]
            # Then search the corresponding index in the diffused block
            slice_in_cell_d = np.searchsorted(smooth_ao_idx, slice_in_cell)

            # j3c_dd may be an h5 object. Load j3c_dd to memory
            d0, d1 = slice_in_cell_d[:2]
            j3c_dd = np.asarray(j3c_dd[d0:d1], order='C')
            naux = j3c_dd.shape[-1]

            drv(j3c.ctypes.data_as(ctypes.c_void_p),
                j3c_dd.ctypes.data_as(ctypes.c_void_p),
                smooth_ao_idx.ctypes.data_as(ctypes.c_void_p),
                (ctypes.c_int*4)(*slice_in_cell),
                (ctypes.c_int*4)(*slice_in_cell_d),
                ctypes.c_int(nao), ctypes.c_int(naod), ctypes.c_int(naux))
            return j3c
        return merge

    def recontract(self, dim=1):
        '''Recontract the vector evaluated with the RS-cell to the vector
        associated to the basis of reference cell
        '''
        ao_loc = self.ref_cell.ao_loc
        ao_map = self.get_ao_indices(self.bas_map, ao_loc)
        nao = ao_loc[-1]

        if dim == 1:

            def recontractor(a):
                assert a.ndim == 2
                a = np.asarray(a, order='C')
                ngrids = a.shape[1]
                out = np.zeros((nao, ngrids), dtype=a.dtype)
                idx = np.arange(ngrids, dtype=np.int32)
                return lib.takebak_2d(out, a, ao_map, idx, thread_safe=False)

        elif dim == 2:

            def recontractor(a):
                assert a.ndim == 2
                a = np.asarray(a, order='C')
                out = np.zeros((nao, nao), dtype=a.dtype)
                return lib.takebak_2d(out, a, ao_map, ao_map, thread_safe=False)

        else:
            raise NotImplementedError(f'dim = {dim}')

        return recontractor

    def recontract_1d(self, vec):
        '''Recontract the vector evaluated with the RS-cell to the vector
        associated to the basis of reference cell
        '''
        return self.recontract()(vec)

    def get_ao_type(self):
        '''Assign a label (STEEP_BASIS, LOCAL_BASIS, SMOOTH_BASIS) to each AO function'''
        ao_loc = self.ao_loc
        nao = ao_loc[-1]
        ao_type = np.empty(nao, dtype=int)

        def assign(type_code):
            ao_idx = self.get_ao_indices(self.bas_type == type_code, ao_loc)
            ao_type[ao_idx] = type_code

        assign(STEEP_BASIS)
        assign(LOCAL_BASIS)
        assign(SMOOTH_BASIS)
        return ao_type

    def decontract_basis(self, to_cart=True, aggregate=False):
        pcell, ctr_coeff = self.ref_cell.decontract_basis(to_cart=to_cart,
                                                          aggregate=aggregate)
        pcell = pcell.view(self.__class__)
        pcell.ref_cell = None

        # Set bas_type labels for the primitive basis of decontracted cell
        smooth_mask = self.bas_type == SMOOTH_BASIS
        smooth_exp_thresholds = {}
        for ia, (ib0, ib1) in enumerate(self.aoslice_by_atom()[:,:2]):
            smooth_bas_ids = ib0 + np.where(smooth_mask[ib0:ib1])[0]
            for ib in smooth_bas_ids:
                l = self._bas[ib,gto.ANG_OF]
                nprim = self._bas[ib,gto.NPRIM_OF]
                pexp = self._bas[ib,gto.PTR_EXP]
                smooth_exp_thresholds[(ia, l)] = max(
                    self._env[pexp:pexp+nprim].max(),
                    smooth_exp_thresholds.get((ia, l), 0))

        pcell_ls = pcell._bas[:,gto.ANG_OF]
        pcell_exps = pcell._env[pcell._bas[:,gto.PTR_EXP]]
        pcell_ao_slices = pcell.aoslice_by_atom()
        pcell.bas_type = np.empty(pcell.nbas, dtype=np.int32)
        pcell.bas_type[:] = LOCAL_BASIS
        for (ia, l), exp_cut in smooth_exp_thresholds.items():
            ib0, ib1 = pcell_ao_slices[ia,:2]
            smooth_mask = ((pcell_exps[ib0:ib1] <= exp_cut+1e-8) &
                           (pcell_ls[ib0:ib1] == l))
            pcell.bas_type[ib0:ib1][smooth_mask] = SMOOTH_BASIS

        pcell.bas_map = np.arange(pcell.nbas, dtype=np.int32)
        pcell.sh_loc = np.append(np.arange(pcell.nbas), pcell.nbas).astype(np.int32)
        logger.debug3(pcell, 'decontracted cell bas_type %s', pcell.bas_type)
        logger.debug3(pcell, 'decontracted cell sh_loc %s', pcell.sh_loc)
        return pcell, ctr_coeff

class ExtendedMole(gto.Mole):
    '''An extended Mole object to mimic periodicity'''
    def __init__(self):
        # The cell which used to generate the supmole
        self.rs_cell: _RangeSeparatedCell = None
        self.bvk_kmesh = None
        self.Ls = None
        self.bvkmesh_Ls = None
        # seg_loc maps the shell Id in bvk cell to shell Id in bvk rs-cell.
        # seg2sh maps the shell Id in bvk rs-cell to the shell Id in supmol.
        # Lattice sum range for each bvk cell shell can be obtained
        # (seg2sh[n+1] - seg2sh[n])
        self.seg_loc = None
        self.seg2sh = None
        # whether the basis bas_mask[bvk-cell-id, basis-id, image-id] is
        # needed to reproduce the periodicity
        self.bas_mask = None
        self.precision = None

    @property
    def sh_loc(self):
        # A map for shell in bvk cell to shell Id in supmol
        return self.seg2sh[self.seg_loc]

    @property
    def bas_map(self):
        # A map to assign each basis of supmol._bas the index in
        # [bvk_cell-id, bas-id, image-id]
        return np.where(self.bas_mask.ravel())[0].astype(np.int32)

    @classmethod
    def from_cell(cls, cell, kmesh, rcut=None, verbose=None):
        if rcut is None: rcut = cell.rcut
        log = logger.new_logger(cell, verbose)

        if not isinstance(cell, _RangeSeparatedCell):
            cell = _RangeSeparatedCell.from_cell(cell)

        bvkcell = pbctools.super_cell(cell, kmesh, wrap_around=True)
        Ls = bvkcell.get_lattice_Ls(rcut=rcut)
        Ls = Ls[np.linalg.norm(Ls, axis=1).argsort()]
        bvkmesh_Ls = k2gamma.translation_vectors_for_kmesh(cell, kmesh, True)
        LKs = Ls[:,None,:] + bvkmesh_Ls
        nimgs, bvk_ncells = LKs.shape[:2]
        log.debug1('Generate supmol with rcut = %g nimgs = %d bvk_ncells = %d',
                   rcut, nimgs, bvk_ncells)

        supmol = cls()
        supmol.__dict__.update(cell.__dict__)
        supmol = pbctools._build_supcell_(supmol, cell, LKs.reshape(nimgs*bvk_ncells, 3))
        supmol.rs_cell = cell
        supmol.bvk_kmesh = kmesh
        supmol.Ls = Ls
        supmol.bvkmesh_Ls = bvkmesh_Ls
        bas_mask = np.ones((bvk_ncells, cell.nbas, nimgs), dtype=bool)
        supmol.seg_loc, supmol.seg2sh = supmol.bas_mask_to_segment(cell, bas_mask, verbose)
        supmol.bas_mask = bas_mask
        supmol.precision = cell.precision
        supmol._env[gto.PTR_EXPCUTOFF] = -np.log(cell.precision*1e-4)

        _bas_reordered = supmol._bas.reshape(
            nimgs, bvk_ncells, cell.nbas, gto.BAS_SLOTS).transpose(1,2,0,3)
        supmol._bas = np.asarray(_bas_reordered.reshape(-1, gto.BAS_SLOTS),
                                 dtype=np.int32, order='C')
        return supmol

    def strip_basis(self, rcut):
        rs_cell = self.rs_cell
        dim = rs_cell.dimension
        if dim == 0:
            return self

        # Search the shortest distance to the reference cell for each atom in the supercell.
        atom_coords = self.atom_coords()
        d = np.linalg.norm(atom_coords[:,None] - rs_cell.atom_coords(), axis=2)
        shortest_dist = np.min(d, axis=1)
        bas_dist = shortest_dist[self._bas[:,gto.ATOM_OF]]

        # filter _bas
        nbas0 = self._bas.shape[0]
        if bas_dist.size == self.bas_mask.size:
            bas_dist = bas_dist.reshape(self.bas_mask.shape)
            self.bas_mask = bas_mask = bas_dist < rcut[:,None]
            self._bas = self._bas[bas_mask.ravel()]
        else:
            dr = np.empty(self.bas_mask.shape)
            dr[:] = 1e9
            dr[self.bas_mask] = bas_dist
            bas_mask = dr < rcut[:,None]
            self._bas = self._bas[bas_mask[self.bas_mask]]
            self.bas_mask = bas_mask

        # filter _atm
        atm_ids = np.unique(self._bas[:,gto.ATOM_OF])
        atm_mapping = np.zeros(self._atm.shape[0], dtype=np.int32)
        atm_mapping[atm_ids] = np.arange(atm_ids.size)
        self._atm = self._atm[atm_ids]
        self._bas[:,gto.ATOM_OF] = atm_mapping[self._bas[:,gto.ATOM_OF]]

        nbas1 = self._bas.shape[0]
        logger.debug1(self, 'strip_basis %d to %d ', nbas0, nbas1)
        self.seg_loc, self.seg2sh = self.bas_mask_to_segment(rs_cell, self.bas_mask)
        return self

    def get_ovlp_mask(self, cutoff=None):
        '''integral screening mask for basis product between cell and supmol'''
        rs_cell = self.rs_cell
        supmol = self
        # consider only the most diffused component of a basis
        cell_exps, cell_cs = pbcgto.cell._extract_pgto_params(rs_cell, 'min')
        cell_l = rs_cell._bas[:,gto.ANG_OF]
        cell_bas_coords = rs_cell.atom_coords()[rs_cell._bas[:,gto.ATOM_OF]]

        if cutoff is None:
            theta_ij = cell_exps.min() / 2
            vol = rs_cell.vol
            lattice_sum_factor = max(2*np.pi*rs_cell.rcut/(vol*theta_ij), 1)
            cutoff = rs_cell.precision/lattice_sum_factor * .1
            logger.debug(self, 'Set ft_ao cutoff to %g', cutoff)

        supmol_exps, supmol_cs = pbcgto.cell._extract_pgto_params(supmol, 'min')
        supmol_bas_coords = supmol.atom_coords()[supmol._bas[:,gto.ATOM_OF]]
        supmol_l = supmol._bas[:,gto.ANG_OF]

        aij = cell_exps[:,None] + supmol_exps
        theta = cell_exps[:,None] * supmol_exps / aij
        dr = np.linalg.norm(cell_bas_coords[:,None,:] - supmol_bas_coords, axis=2)

        aij1 = 1./aij
        aij2 = aij**-.5
        dri = supmol_exps*aij1 * dr + aij2
        drj = cell_exps[:,None]*aij1 * dr + aij2
        norm_i = cell_cs * ((2*cell_l+1)/(4*np.pi))**.5
        norm_j = supmol_cs * ((2*supmol_l+1)/(4*np.pi))**.5
        fl = 2*np.pi/rs_cell.vol*dr/theta + 1.
        ovlp = (np.pi**1.5 * norm_i[:,None]*norm_j * np.exp(-theta*dr**2) *
                dri**cell_l[:,None] * drj**supmol_l * aij1**1.5 * fl)
        return ovlp > cutoff

    @staticmethod
    def bas_mask_to_segment(rs_cell, bas_mask, verbose=None):
        '''
        bas_mask shape [bvk_ncells, nbas, nimgs]
        '''
        log = logger.new_logger(rs_cell, verbose)
        bvk_ncells, cell_rs_nbas, nimgs = bas_mask.shape
        images_count = np.count_nonzero(bas_mask, axis=2)

        # seg_loc maps shell Id in bvk-cell to segment Id in supmol
        # seg2sh maps the segment Id to shell Id of supmol
        seg_loc = np.arange(bvk_ncells)[:,None] * cell_rs_nbas + rs_cell.sh_loc[:-1]
        seg_loc = np.append(seg_loc.ravel(), bvk_ncells * cell_rs_nbas)
        seg2sh = np.append(0, np.cumsum(images_count.ravel()))

        if log.verbose > logger.DEBUG:
            steep_mask = rs_cell.bas_type == STEEP_BASIS
            local_mask = rs_cell.bas_type == LOCAL_BASIS
            diffused_mask = rs_cell.bas_type == SMOOTH_BASIS
            log.debug1('No. steep basis in sup-mol %d', images_count[:,steep_mask].sum())
            log.debug1('No. local basis in sup-mol %d', images_count[:,local_mask].sum())
            log.debug1('No. diffused basis in sup-mol %d', images_count[:,diffused_mask].sum())
            log.debug3('sup-mol seg_loc %s', seg_loc)
            log.debug3('sup-mol seg2sh %s', seg2sh)
        return seg_loc.astype(np.int32), seg2sh.astype(np.int32)

    def bas_type_to_indices(self, type_code=SMOOTH_BASIS):
        '''Return the basis indices of required bas_type'''
        cell0_mask = self.rs_cell.bas_type == type_code
        if np.any(cell0_mask):
            # (bvk_ncells, rs_cell.nbas, nimgs)
            bas_type_mask = np.empty_like(self.bas_mask)
            bas_type_mask[:] = cell0_mask[None,:,None]
            bas_type_mask = bas_type_mask[self.bas_mask]
            return np.where(bas_type_mask)[0]
        else:
            return np.arange(0)

    gen_ft_kernel = gen_ft_kernel

def estimate_rcut(cell, precision=None):
    '''Estimate rcut for each basis based on Schwarz inequality
    Q_ij ~ S_ij * (sqrt(2aij/pi) * aij**(lij*2) * (4*lij-1)!!)**.5
    '''
    if precision is None:
        # The rcut estimated with this function is sufficient to converge
        # the integrals to the required precision. Errors around the required
        # precision is found when checking hermitian symmetry of the integrals.
        # The discrepancy in hermitian symmetry may cause issues in post-HF
        # methods which assume the hermitian symmetry in MO integrals.
        # Therefore precision is adjusted to ensure hermitian symmetry.
        precision = cell.precision * 1e-2

    if cell.nbas == 0:
        return np.zeros(1)

    # consider only the most diffused component of a basis
    exps, cs = pbcgto.cell._extract_pgto_params(cell, 'min')
    ls = cell._bas[:,gto.ANG_OF]

    ai_idx = exps.argmin()
    ai = exps[ai_idx]
    li = ls[ai_idx]
    ci = cs[ai_idx]
    aj = exps
    lj = ls
    cj = cs
    aij = ai + aj
    lij = li + lj
    norm_ang = ((2*li+1)*(2*lj+1))**.5/(4*np.pi)
    c1 = ci * cj * norm_ang
    theta = ai * aj / aij
    aij1 = aij**-.5
    fac = np.pi**1.5*c1 * aij1**(lij+3) * (2*aij/np.pi)**.25 * aij**lij
    fac /= precision

    r0 = cell.rcut
    dri = aj*aij1 * r0 + 1.
    drj = ai*aij1 * r0 + 1.
    fl = 2*np.pi/cell.vol * r0/theta
    r0 = (np.log(fac * dri**li * drj**lj * fl + 1.) / theta)**.5

    dri = aj*aij1 * r0 + 1.
    drj = ai*aij1 * r0 + 1.
    fl = 2*np.pi/cell.vol * r0/theta
    r0 = (np.log(fac * dri**li * drj**lj * fl + 1.) / theta)**.5
    return r0



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#         Peter Koval <koval.peter@gmail.com>
#         Paul J. Robinson <pjrobinson@ucla.edu>
#

'''
Gaussian cube file format.  Reference:
http://paulbourke.net/dataformats/cube/
https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html
http://gaussian.com/cubegen/

The output cube file has the following format

Comment line
Comment line
N_atom Ox Oy Oz         # number of atoms, followed by the coordinates of the origin
N1 vx1 vy1 vz1          # number of grids along each axis, followed by the step size in x/y/z direction.
N2 vx2 vy2 vz2          # ...
N3 vx3 vy3 vz3          # ...
Atom1 Z1 x y z          # Atomic number, charge, and coordinates of the atom
...                     # ...
AtomN ZN x y z          # ...
Data on grids           # (N1*N2) lines of records, each line has N3 elements
'''

import time
import numpy
import pyscf
from pyscf import lib
from pyscf import gto
from pyscf import df
from pyscf.dft import numint
from pyscf import __config__

RESOLUTION = getattr(__config__, 'cubegen_resolution', None)
BOX_MARGIN = getattr(__config__, 'cubegen_box_margin', 3.0)
ORIGIN = getattr(__config__, 'cubegen_box_origin', None)
# If given, EXTENT should be a 3-element ndarray/list/tuple to represent the
# extension in x, y, z
EXTENT = getattr(__config__, 'cubegen_box_extent', None)


def density(mol, outfile, dm, nx=80, ny=80, nz=80, resolution=RESOLUTION,
            margin=BOX_MARGIN):
    """Calculates electron density and write out in cube format.

    Args:
        mol : Mole
            Molecule to calculate the electron density for.
        outfile : str
            Name of Cube file to be written.
        dm : ndarray
            Density matrix of molecule.

    Kwargs:
        nx : int
            Number of grid point divisions in x direction.
            Note this is function of the molecule's size; a larger molecule
            will have a coarser representation than a smaller one for the
            same value. Conflicts to keyword resolution.
        ny : int
            Number of grid point divisions in y direction.
        nz : int
            Number of grid point divisions in z direction.
        resolution: float
            Resolution of the mesh grid in the cube box. If resolution is
            given in the input, the input nx/ny/nz have no effects.  The value
            of nx/ny/nz will be determined by the resolution and the cube box
            size.
    """
    from pyscf.pbc.gto import Cell
    cc = Cube(mol, nx, ny, nz, resolution, margin)

    GTOval = 'GTOval'
    if isinstance(mol, Cell):
        GTOval = 'PBC' + GTOval

    # Compute density on the .cube grid
    coords = cc.get_coords()
    ngrids = cc.get_ngrids()
    blksize = min(8000, ngrids)
    rho = numpy.empty(ngrids)
    for ip0, ip1 in lib.prange(0, ngrids, blksize):
        ao = mol.eval_gto(GTOval, coords[ip0:ip1])
        rho[ip0:ip1] = numint.eval_rho(mol, ao, dm)
    rho = rho.reshape(cc.nx,cc.ny,cc.nz)

    # Write out density to the .cube file
    cc.write(rho, outfile, comment='Electron density in real space (e/Bohr^3)')
    return rho


def orbital(mol, outfile, coeff, nx=80, ny=80, nz=80, resolution=RESOLUTION,
            margin=BOX_MARGIN):
    """Calculate orbital value on real space grid and write out in cube format.

    Args:
        mol : Mole
            Molecule to calculate the electron density for.
        outfile : str
            Name of Cube file to be written.
        coeff : 1D array
            coeff coefficient.

    Kwargs:
        nx : int
            Number of grid point divisions in x direction.
            Note this is function of the molecule's size; a larger molecule
            will have a coarser representation than a smaller one for the
            same value. Conflicts to keyword resolution.
        ny : int
            Number of grid point divisions in y direction.
        nz : int
            Number of grid point divisions in z direction.
        resolution: float
            Resolution of the mesh grid in the cube box. If resolution is
            given in the input, the input nx/ny/nz have no effects.  The value
            of nx/ny/nz will be determined by the resolution and the cube box
            size.
    """
    from pyscf.pbc.gto import Cell
    cc = Cube(mol, nx, ny, nz, resolution, margin)

    GTOval = 'GTOval'
    if isinstance(mol, Cell):
        GTOval = 'PBC' + GTOval

    # Compute density on the .cube grid
    coords = cc.get_coords()
    ngrids = cc.get_ngrids()
    blksize = min(8000, ngrids)
    orb_on_grid = numpy.empty(ngrids)
    for ip0, ip1 in lib.prange(0, ngrids, blksize):
        ao = mol.eval_gto(GTOval, coords[ip0:ip1])
        orb_on_grid[ip0:ip1] = numpy.dot(ao, coeff)
    orb_on_grid = orb_on_grid.reshape(cc.nx,cc.ny,cc.nz)

    # Write out orbital to the .cube file
    cc.write(orb_on_grid, outfile, comment='Orbital value in real space (1/Bohr^3)')
    return orb_on_grid


def mep(mol, outfile, dm, nx=80, ny=80, nz=80, resolution=RESOLUTION,
        margin=BOX_MARGIN):
    """Calculates the molecular electrostatic potential (MEP) and write out in
    cube format.

    Args:
        mol : Mole
            Molecule to calculate the electron density for.
        outfile : str
            Name of Cube file to be written.
        dm : ndarray
            Density matrix of molecule.

    Kwargs:
        nx : int
            Number of grid point divisions in x direction.
            Note this is function of the molecule's size; a larger molecule
            will have a coarser representation than a smaller one for the
            same value. Conflicts to keyword resolution.
        ny : int
            Number of grid point divisions in y direction.
        nz : int
            Number of grid point divisions in z direction.
        resolution: float
            Resolution of the mesh grid in the cube box. If resolution is
            given in the input, the input nx/ny/nz have no effects.  The value
            of nx/ny/nz will be determined by the resolution and the cube box
            size.
    """
    cc = Cube(mol, nx, ny, nz, resolution, margin)

    coords = cc.get_coords()

    # Nuclear potential at given points
    Vnuc = 0
    for i in range(mol.natm):
        r = mol.atom_coord(i)
        Z = mol.atom_charge(i)
        rp = r - coords
        Vnuc += Z / numpy.einsum('xi,xi->x', rp, rp)**.5

    # Potential of electron density
    Vele = numpy.empty_like(Vnuc)
    for p0, p1 in lib.prange(0, Vele.size, 600):
        fakemol = gto.fakemol_for_charges(coords[p0:p1])
        ints = df.incore.aux_e2(mol, fakemol)
        Vele[p0:p1] = numpy.einsum('ijp,ij->p', ints, dm)

    MEP = Vnuc - Vele     # MEP at each point
    MEP = MEP.reshape(cc.nx,cc.ny,cc.nz)

    # Write the potential
    cc.write(MEP, outfile, 'Molecular electrostatic potential in real space')
    return MEP


class Cube:
    '''  Read-write of the Gaussian CUBE files

    Attributes:
        nx : int
            Number of grid point divisions in x direction.
            Note this is function of the molecule's size; a larger molecule
            will have a coarser representation than a smaller one for the
            same value. Conflicts to keyword resolution.
        ny : int
            Number of grid point divisions in y direction.
        nz : int
            Number of grid point divisions in z direction.
        resolution: float
            Resolution of the mesh grid in the cube box. If resolution is
            given in the input, the input nx/ny/nz have no effects.  The value
            of nx/ny/nz will be determined by the resolution and the cube box
            size. The unit is Bohr.
    '''
    def __init__(self, mol, nx=80, ny=80, nz=80, resolution=RESOLUTION,
                 margin=BOX_MARGIN, origin=ORIGIN, extent=EXTENT):
        from pyscf.pbc.gto import Cell
        self.mol = mol
        coord = mol.atom_coords()

        # If the molecule is periodic, use lattice vectors as the box
        # and automatically determine margin, origin, and extent
        if isinstance(mol, Cell):
            self.box = mol.lattice_vectors()
            atom_center = (numpy.max(coord, axis=0) + numpy.min(coord, axis=0))/2
            box_center = (self.box[0] + self.box[1] + self.box[2])/2
            self.boxorig = atom_center - box_center
        else:
            # Create a box based on its origin and extents/lengths (rectangular cuboid only)
            # If extent is not supplied, use the coordinates plus a margin on both sides
            if extent is None:
                extent = numpy.max(coord, axis=0) - numpy.min(coord, axis=0) + 2*margin
            self.box = numpy.diag(extent)

            # if origin is not supplied, set it as the minimum coordinate minus a margin.
            if origin is None:
                origin = numpy.min(coord, axis=0) - margin
            self.boxorig = numpy.asarray(origin)

        if resolution is not None:
            nx, ny, nz = numpy.ceil(numpy.diag(self.box) / resolution).astype(int)

        self.nx = nx
        self.ny = ny
        self.nz = nz

        if isinstance(mol, Cell):
            # Use an asymmetric mesh for tiling unit cells
            self.xs = numpy.linspace(0, 1, nx, endpoint=False)
            self.ys = numpy.linspace(0, 1, ny, endpoint=False)
            self.zs = numpy.linspace(0, 1, nz, endpoint=False)
        else:
            # Use endpoint=True to get a symmetric mesh
            # see also the discussion https://github.com/sunqm/pyscf/issues/154
            self.xs = numpy.linspace(0, 1, nx, endpoint=True)
            self.ys = numpy.linspace(0, 1, ny, endpoint=True)
            self.zs = numpy.linspace(0, 1, nz, endpoint=True)

    def get_coords(self) :
        """  Result: set of coordinates to compute a field which is to be stored
        in the file.
        """
        frac_coords = lib.cartesian_prod([self.xs, self.ys, self.zs])
        return frac_coords @ self.box + self.boxorig # Convert fractional coordinates to real-space coordinates

    def get_ngrids(self):
        return self.nx * self.ny * self.nz

    def get_volume_element(self):
        return (self.xs[1]-self.xs[0])*(self.ys[1]-self.ys[0])*(self.zs[1]-self.zs[0])

    def write(self, field, fname, comment=None):
        """  Result: .cube file with the field in the file fname.  """
        assert (field.ndim == 3)
        assert (field.shape == (self.nx, self.ny, self.nz))
        if comment is None:
            comment = 'Generic field? Supply the optional argument "comment" to define this line'

        mol = self.mol
        coord = mol.atom_coords()
        with open(fname, 'w') as f:
            f.write(comment+'\n')
            f.write(f'PySCF Version: {pyscf.__version__}  Date: {time.ctime()}\n')
            f.write(f'{mol.natm:5d}')
            f.write('%12.6f%12.6f%12.6f\n' % tuple(self.boxorig.tolist()))
            dx = self.xs[-1] if len(self.xs) == 1 else self.xs[1]
            dy = self.ys[-1] if len(self.ys) == 1 else self.ys[1]
            dz = self.zs[-1] if len(self.zs) == 1 else self.zs[1]
            delta = (self.box.T * [dx,dy,dz]).T
            f.write(f'{self.nx:5d}{delta[0,0]:12.6f}{delta[0,1]:12.6f}{delta[0,2]:12.6f}\n')
            f.write(f'{self.ny:5d}{delta[1,0]:12.6f}{delta[1,1]:12.6f}{delta[1,2]:12.6f}\n')
            f.write(f'{self.nz:5d}{delta[2,0]:12.6f}{delta[2,1]:12.6f}{delta[2,2]:12.6f}\n')
            for ia in range(mol.natm):
                atmsymb = mol.atom_symbol(ia)
                f.write('%5d%12.6f'% (gto.charge(atmsymb), 0.))
                f.write('%12.6f%12.6f%12.6f\n' % tuple(coord[ia]))

            for ix in range(self.nx):
                for iy in range(self.ny):
                    for iz0, iz1 in lib.prange(0, self.nz, 6):
                        fmt = '%13.5E' * (iz1-iz0) + '\n'
                        f.write(fmt % tuple(field[ix,iy,iz0:iz1].tolist()))

    def read(self, cube_file):
        with open(cube_file, 'r') as f:
            f.readline()
            f.readline()
            data = f.readline().split()
            natm = int(data[0])
            self.boxorig = numpy.array([float(x) for x in data[1:]])
            def parse_nx(data):
                from pyscf.pbc.gto import Cell
                d = data.split()
                nx = int(d[0])
                x_vec = numpy.array([float(x) for x in d[1:]]) * nx
                if isinstance(self.mol, Cell):
                    # Use an asymmetric mesh for tiling unit cells
                    xs = numpy.linspace(0, 1, nx, endpoint=False)
                else:
                    # Use endpoint=True to get a symmetric mesh
                    # see also the discussion https://github.com/sunqm/pyscf/issues/154
                    xs = numpy.linspace(0, 1, nx, endpoint=True)
                return x_vec, nx, xs
            self.box = numpy.zeros((3,3))
            self.box[0], self.nx, self.xs = parse_nx(f.readline())
            self.box[1], self.ny, self.ys = parse_nx(f.readline())
            self.box[2], self.nz, self.zs = parse_nx(f.readline())
            atoms = []
            for ia in range(natm):
                d = f.readline().split()
                atoms.append([int(d[0]), [float(x) for x in d[2:]]])
            self.mol = gto.M(atom=atoms, unit='Bohr')

            data = f.read()
        cube_data = numpy.array([float(x) for x in data.split()])
        return cube_data.reshape([self.nx, self.ny, self.nz])


if __name__ == '__main__':
    from pyscf import scf
    from pyscf.tools import cubegen
    mol = gto.M(atom='''O 0.00000000,  0.000000,  0.000000
                H 0.761561, 0.478993, 0.00000000
                H -0.761561, 0.478993, 0.00000000''', basis='6-31g*')
    mf = scf.RHF(mol).run()
    cubegen.density(mol, 'h2o_den.cube', mf.make_rdm1()) #makes total density
    cubegen.mep(mol, 'h2o_pot.cube', mf.make_rdm1())
    cubegen.orbital(mol, 'h2o_mo1.cube', mf.mo_coeff[:,0])



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import copy
from scipy import linalg
from pyscf import lib, dft
from pyscf.lib import logger
from pyscf.dft import libxc
from pyscf.dft.gen_grid import Grids
from pyscf.dft.numint import _NumInt, NumInt
from pyscf.mcpdft import pdft_veff, tfnal_derivs, _libxc, _dms, pdft_feff, pdft_eff
from pyscf.mcpdft.otpd import get_ontop_pair_density
from pyscf import __config__

FT_R0 = getattr(__config__, 'mcpdft_otfnal_ftransfnal_R0', 0.9)
FT_R1 = getattr(__config__, 'mcpdft_otfnal_ftransfnal_R1', 1.15)
FT_A = getattr(__config__, 'mcpdft_otfnal_ftransfnal_A', -475.60656009)
FT_B = getattr(__config__, 'mcpdft_otfnal_ftransfnal_B', -379.47331922)
FT_C = getattr(__config__, 'mcpdft_otfnal_ftransfnal_C', -85.38149682)

OT_ALIAS = {'MC23': 'tMC23'}
OT_HYB_ALIAS = {'PBE0' : '0.25*HF + 0.75*PBE, 0.25*HF + 0.75*PBE',
                }

REG_OT_FUNCTIONALS={}

# ALIAS for the preset on-top functional
OT_PRESET={
    # Reparametrized-M06L: rep-M06L
    # MC23 = { '0.2952*HF + (1-0.2952)*rep-M06L, 0.2952*HF + (1-0.2952)*rep-M06L'}}
    # XC_ID_MGGA_C_M06_L = 233
    # XC_ID_MGGA_X_M06_L = 203
    'MC23':{
        'xc_base':'M06L',
        'ext_params':{203: np.array([3.352197, 6.332929e-01, -9.469553e-01, 2.030835e-01,
                                     2.503819, 8.085354e-01, -3.619144, -5.572321e-01,
                                     -4.506606, 9.614774e-01, 6.977048, -1.309337, -2.426371,
                                     -7.896540e-03, 1.364510e-02, -1.714252e-06, -4.698672e-05, 0.0]),
                        233: np.array([0.06, 0.0031, 0.00515088, 0.00304966, 2.427648, 3.707473,
                                       -7.943377, -2.521466, 2.658691, 2.932276, -8.832841e-01,
                                       -1.895247, -2.899644, -5.068570e-01, -2.712838, 9.416102e-02,
                                       -3.485860e-03, -5.811240e-04, 6.668814e-04, 0.0, 2.669169e-01,
                                       -7.563289e-02, 7.036292e-02, 3.493904e-04, 6.360837e-04, 0.0, 1e-10])},
        'hyb':(0.2952,0.2952,0),
        'facs':(0.7048,0.7048)}
        }

def register_otfnal(xc_code, preset):
    '''
    This function registers the new on-top functional if it hasn't been
    registered previously.
    Args:
        xc_code: str
            The name of the on-top functional to be registered.
        preset: dict
            The dictionary containing the information about the on-top functional
            to be registered.
            xc_base: str
                The name of the underylying KS-functional in the libxc library.
            ext_params: dict, with LibXC exchange and correlation functional integer ID as key, and
                an array-like object containing the functional parameters as value.
            hyb: tuple
                The hybrid functional parameters.
            facs: tuple
                The mixing factors.
            kwargs: dict
                The additional keyword arguments.
    '''
    libxc_register_code = xc_code.upper ()
    libxc_base_code = preset['xc_base']
    ext_params = preset['ext_params']
    hyb = preset.get('hyb', None)
    facs = preset.get('facs', None)
    libxc.register_custom_functional_(libxc_register_code, libxc_base_code,
                                      ext_params=ext_params, hyb=hyb, facs=facs)
    REG_OT_FUNCTIONALS[xc_code.upper()] = {'hyb_x':preset.get('hyb',[0])[0],
                                           'hyb_c':preset.get('hyb',[0])[0]}

def unregister_otfnal(xc_code):
    '''
    This function unregisters the on-top functional if it has been registered
    previously.
    Args:
        xc_code: str
            The name of the on-top functional to be unregistered.
    '''
    try:
        if xc_code.upper() in REG_OT_FUNCTIONALS:
            libxc_unregister_code = xc_code.upper()
            libxc.unregister_custom_functional_(libxc_unregister_code)
            del REG_OT_FUNCTIONALS[xc_code.upper()]

    except Exception as e:
        raise RuntimeError(f"Failed to unregister functional '{xc_code}': {e}") from e

def _get_registered_ot_functional(xc_code, mol):
    '''
    This function returns the on-top functional if it has been registered
    previously.
    Args:
        xc_code: str
            The name of the on-top functional to be registered.
    '''
    if (xc_code.upper() not in REG_OT_FUNCTIONALS) and (xc_code.upper() in OT_PRESET):
        preset = OT_PRESET[xc_code.upper()]
        register_otfnal(xc_code, preset)
        logger.info(mol, 'Registered the on-top functional: %s', xc_code)
    return xc_code.upper()

def energy_ot (ot, casdm1s, casdm2, mo_coeff, ncore, max_memory=2000, hermi=1):
    '''Compute the on-top energy - the last term in

    E_MCPDFT = h_pq l_pq + 1/2 v_pqrs l_pq l_rs + E_ot[rho,Pi]

    Args:
        ot : an instance of otfnal class
        casdm1s : ndarray of shape (2, ncas, ncas)
            Contains spin-separated one-body density matrices in an
            active-orbital basis
        casdm2 : ndarray of shape (ncas, ncas, ncas, ncas)
            Contains spin-summed two-body density matrix in an active-
            orbital basis
        mo_coeff : ndarray of shape (nao, nmo)
            Contains molecular orbital coefficients for active-space
            orbitals. Columns ncore through ncore+ncas give the basis
            in which casdm1s and casdm2 are expressed.
        ncore : integer
            Number of doubly occupied inactive "core" orbitals not
            explicitly included in casdm1s and casdm2

    Kwargs:
        max_memory : int or float
            maximum cache size in MB
            default is 2000
        hermi : int
            1 if 1rdms are assumed hermitian, 0 otherwise

    Returns : float
        The MC-PDFT on-top (nonclassical) energy
    '''
    E_ot = 0.0
    ni, xctype = ot._numint, ot.xctype
    if xctype=='HF': return E_ot
    dens_deriv = ot.dens_deriv
    Pi_deriv = ot.Pi_deriv

    nao = mo_coeff.shape[0]
    ncas = casdm2.shape[0]
    cascm2 = _dms.dm2_cumulant (casdm2, casdm1s)
    dm1s = _dms.casdm1s_to_dm1s (ot, casdm1s, mo_coeff=mo_coeff, ncore=ncore,
                                 ncas=ncas)
    mo_cas = mo_coeff[:,ncore:][:,:ncas]

    t0 = (logger.process_clock (), logger.perf_counter ())
    make_rho = tuple (ni._gen_rho_evaluator (ot.mol, dm1s[i,:,:], hermi=hermi, with_lapl=False) for
        i in range(2))
    for ao, mask, weight, _ in ni.block_loop (ot.mol, ot.grids, nao,
            dens_deriv, max_memory):
        rho = np.asarray ([m[0] (0, ao, mask, xctype) for m in make_rho])
        t0 = logger.timer (ot, 'untransformed density', *t0)
        Pi = get_ontop_pair_density (ot, rho, ao, cascm2, mo_cas,
            Pi_deriv, mask)
        t0 = logger.timer (ot, 'on-top pair density calculation', *t0)
        if rho.ndim == 2:
            rho = np.expand_dims (rho, 1)
            Pi = np.expand_dims (Pi, 0)
        E_ot += ot.eval_ot (rho, Pi, dderiv=0, weights=weight)[0].dot (weight)
        t0 = logger.timer (ot, 'on-top energy calculation', *t0)

    return E_ot

class otfnal:
    r''' Parent class of on-top pair-density functional. The main
    callable is ``eval_ot,'' which is comparable to pyscf.dft.libxc
    ``eval_xc.'' A true ``kernel'' method, which would take arbitrary
    1- and 2-RDMs and return the total PDFT energy, awaits design
    decisions on how far I'm willing/able to generalize the otpd
    functions. For instance, in MP2 or CCSD, the 2-RDM spans the
    whole orbital space and it may not be possible to hold it in
    memory. At present, it's all designed around MC-SCF, which is
    why the ``kernel'' function that actually calculates the energy
    is in mcpdft.py instead of here.

    Attributes:
        mol : object of class pyscf.gto.mole
        grids : object of class pyscf.dft.gen_grid.Grids
        eval_ot : function with calling signature shown below
        _numint : object of class pyscf.dft.NumInt
            member functions "hybrid_coeff", "nlc_coeff, "rsh_coeff",
            and "_xc_type" (at least) must be overloaded; see below
        otxc : string
            name of on-top pair-density exchange-correlation functional
    '''

    def __init__ (self, mol):
        self.mol = mol
        self.verbose = mol.verbose
        self.stdout = mol.stdout

    Pi_deriv = 0

    def _init_info (self):
        logger.info (self, 'Building %s functional', self.otxc)
        hyb = self._numint.rsh_and_hybrid_coeff(self.otxc, spin=self.mol.spin)[2]
        if hyb[0] > 0:
            logger.info (self, 'Hybrid functional with %s CASSCF exchange',
                hyb)

    @property
    def xctype (self):
        return self._numint._xc_type (self.otxc)

    @property
    def dens_deriv (self):
        return ['LDA', 'GGA', 'MGGA'].index (self.xctype)

    def eval_ot (self, rho, Pi, dderiv=0, **kwargs):
        r''' Evaluate the on-dop energy and its functional derivatives
             on a grid

        Args:
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]

        Kwargs:
            dderiv : integer
                Order of derivatives to return

        Returns:
            eot : ndarray of shape (ngrids)
                integrand of the on-top exchange-correlation energy
            vot : (array_like (rho), array_like (Pi)) or None
                first functional derivative of Eot wrt (density, pair-
                density) and their derivatives
            fot : ndarray of shape (*,ngrids) or None
                second functional derivative of Eot wrt density, pair-
                density, and derivatives; first dimension is lower-
                triangular matrix elements corresponding to the basis
                (rho, Pi, |drho|^2, drho'.dPi, |dPi|) stopping at Pi (3
                elements) for t-LDA and |drho|^2 (6 elements) for t-GGA.
        '''
        raise NotImplementedError("on-top xc functional not defined")

    energy_ot = energy_ot
    get_eff_1body = pdft_eff.get_eff_1body
    get_eff_2body = pdft_eff.get_eff_2body
    get_eff_2body_kl = pdft_eff.get_eff_2body_kl

    get_veff_1body = pdft_veff.get_veff_1body
    get_veff_2body = pdft_veff.get_veff_2body
    get_veff_2body_kl = pdft_veff.get_veff_2body_kl

    get_feff_1body = pdft_feff.get_feff_1body
    get_feff_2body = pdft_feff.get_feff_2body

    def reset (self, mol=None):
        ''' Discard cached grid data and optionally update the mol '''
        if mol is not None:
            self.mol = mol
        self.grids.reset (mol=mol)

class transfnal (otfnal):
    __doc__ = otfnal.__doc__ + r'''

    ``translated functional'' of Li Manni et al., JCTC 10, 3669 (2014).
    The extra attributes are all callables; see their docstrings for
    more information.

    Args:
        ks : object of :class:`dft.RKS`
            ks.xc is the Kohn-Sham functional being ``translated''

    '''

    transl_prefix='t'

    def __init__ (self, ks, **kwargs):
        otfnal.__init__(self, ks.mol, **kwargs)
        self.otxc = 't' + ks.xc
        self._numint = copy.copy (ks._numint)
        self._numint.libxc = libxc
        self.grids = copy.copy (ks.grids)
        self._numint.hybrid_coeff = t_hybrid_coeff.__get__(self._numint)
        self._numint.nlc_coeff = t_nlc_coeff.__get__(self._numint)
        self._numint.rsh_coeff = t_rsh_coeff.__get__(self._numint)
        self._numint.eval_xc = t_eval_xc.__get__(self._numint)
        self._numint._xc_type = t_xc_type.__get__(self._numint)
        self._init_info ()

    def get_ratio (self, Pi, rho_avg):
        r''' R = Pi / [rho/2]^2 = Pi / rho_avg^2
            An intermediate quantity when computing the translated spin
            densities

            Note this function returns 1 for values and 0 for
            derivatives for every point where the charge density is
            close to zero (i.e., convention: 0/0 = 1)

            Args:
                Pi : ndarray of shape (*,ngrids)
                    Contains on-top pair density on a grid
                rho_avg : ndarray of shape (*,ngrids)
                    Contains the average of the spin-up and spin-down
                    charge densities on a grid, (rho[0]+rho[1])/2

            Returns:
                R : ndarray of shape (*,ngrids)
                    on-top ratio
        '''
        nderiv = min (rho_avg.shape[0], Pi.shape[0])
        ngrids = rho_avg.shape[1]
        assert (Pi.shape[1] == ngrids)
        if nderiv > 4:
            raise NotImplementedError("derivatives above order 1")

        R = np.zeros ((nderiv,ngrids), dtype=Pi.dtype)
        R[0,:] = 1
        R[0, Pi[0] == 0] = 0.0
        idx = rho_avg[0] >= (1e-15 / 2)
        # Chain rule!
        for ideriv in range(nderiv):
            R[ideriv,idx] = Pi[ideriv,idx] / rho_avg[0,idx] / rho_avg[0,idx]

        # Product rule!
        for ideriv in range (1,nderiv):
            R[ideriv,idx] -= (2 * rho_avg[ideriv,idx] * R[0,idx]
                / rho_avg[0,idx])
        return R

    def get_rho_translated (self, Pi, rho, _fn_deriv=0):
        r''' Compute the "translated" alpha and beta densities:
        For the unrestricted case,
        rho = [rho^a, rho^b]
        Here:
            rho^a will have dim of 1,4 or 6 depends on the functional. For MGGA,
            rho^a = [rho_u,grad_xu, grad_yu, grad_zu, laplacian_u, tau_u]
            Similar for rho_b.

        The translation is done as follows:

        rho_t^a = (rho/2) * (1 + zeta)
        rho_t^b = (rho/2) * (1 - zeta)
        rho'_t^a = (rho'/2) * (1 + zeta)
        rho'_t^b = (rho'/2) * (1 - zeta)
        tau_t^a = (tau/2) * (1 + zeta)
        tau_t^b = (tau/2) * (1 - zeta)

        See "get_zeta" for the meaning of "zeta"

        Args:
            Pi : ndarray of shape (*, ngrids)
                containing on-top pair density [and derivatives]
            rho : ndarray of shape (2, *, ngrids)
                containing spin density [and derivatives]

        Kwargs:
            _fn_deriv : integer
                Order of functional derivatives of zeta to compute.
                In "translated" functionals, no functional derivatives
                of zeta are used. This kwarg is used for convenience
                when calling from children classes. It changes the
                return signature and should not normally be touched by
                users.

        Returns:
            rho_t : ndarray of shape (2,*,ngrids)
                Translated spin density (and derivatives) in case of LDA or GGAs
                Translated spin density, derivatives, and kinetic energy density in case of MGGA
        '''

        # For nonzero charge & pair density, set alpha dens = beta dens
        # = 1/2 charge dens
        rho_avg = (rho[0,:,:] + rho[1,:,:]) / 2
        rho_t = rho.copy ()
        rho_t[0] = rho_t[1] = rho_avg

        # For 0 <= ratio < 1 and 0 <= rho, correct spin density using on-top
        # density
        nderiv_R = Pi.shape[0] if _fn_deriv else 1
        R = self.get_ratio (Pi[0:nderiv_R,:], rho_avg[0:nderiv_R,:])
        zeta = self.get_zeta (R, fn_deriv=_fn_deriv)

        # Chain rule!
        w = rho_avg * zeta[0:1]
        rho_t[0] += w
        rho_t[1] -= w

        if _fn_deriv > 0: return rho_t, R, zeta
        return rho_t

    def get_zeta (self, R, fn_deriv=0, _Rmax=1):
        r''' Compute the intermediate zeta used to compute the
        translated spin densities and its functional derivatives

        From the original translation [Li Manni et al., JCTC 10, 3669
        (2014)]:

        zeta = (1-ratio)^(1/2) ; ratio < 1
             = 0               ; otherwise

        Args:
            R : ndarray of shape (*,ngrids)
                Ratio (4Pi/rho^2) and possibly its spatial derivatives
                Only the first row is used in this function

        Kwargs:
            fn_deriv : integer
                order of functional derivative (d^n z / dR^n) to return
                along with the value of zeta
            _Rmax : float
                maximum value of R for which to compute zeta or its
                derivatives; columns of zeta with R[0]>_Rmax are zero.
                This is a hook for the ``fully-translated'' child class
                and should not be touched normally.

        Returns:
            zeta : ndarray of shape (fn_deriv+1, ngrids)
        '''
        if R.ndim == 2: R = R[0]
        ngrids = R.size
        zeta = np.zeros ((fn_deriv+1, ngrids), dtype=R.dtype)
        idx = R < _Rmax
        zeta[0,idx] = np.sqrt (1.0 - R[idx])
        if fn_deriv:
            zeta[1,idx] = -0.5 / zeta[0,idx]
        if fn_deriv > 1: fac = 0.5 / (1.0-R[idx])
        for n in range (1,fn_deriv):
            zeta[n+1,idx] = zeta[n,idx] * (2*n-1) * fac
        return zeta

    def split_x_c (self):
        ''' Get one translated functional for just the exchange and one
            for just the correlation part of the energy.

            Returns:
                xfnal : object of :class:`transfnal`
                    this functional, but only the exchange part
                cfnal : object of :class:`transfnal`
                    this functional, but only the correlation part
        '''
        xc_base = self.otxc[len (self.transl_prefix):]
        x_code, c_code = _libxc.split_x_c_comma (xc_base)
        x_code = self.transl_prefix + x_code + ','
        c_code = self.transl_prefix + ',' + c_code
        xfnal = copy.copy (self)
        xfnal._numint = copy.copy (self._numint)
        xfnal.grids = copy.copy (self.grids)
        xfnal.verbose = self.verbose
        xfnal.stdout = self.stdout
        xfnal.otxc = x_code
        cfnal = copy.copy (self)
        cfnal._numint = copy.copy (self._numint)
        cfnal.grids = copy.copy (self.grids)
        cfnal.verbose = self.verbose
        cfnal.stdout = self.stdout
        cfnal.otxc = c_code
        return xfnal, cfnal

    def jT_op (self, x, rho, Pi):
        r''' Evaluate jTx = (x.j)T where j is the Jacobian of the
        translated densities in terms of the untranslated density and
        pair density

        Args:
            x : ndarray of shape (2,*,ngrids)
                Usually, a functional derivative of the on-top xc energy
                wrt translated densities
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]

        Returns: ndarray of shape (*,ngrids)
            Usually, a functional derivative of the on-top pair density
            exchange-correlation energy wrt to total density and its
            derivatives. The potential must be spin-symmetric in
            pair-density functional theory.
            2 rows for tLDA, 3 rows for tGGA, and 4 rows for meta-GGA
        '''
        # ordering: rho, Pi, |rho'|^2, tau
        ncol = (2, 3, 4)[self.dens_deriv]
        ngrid = rho.shape[-1]
        jTx = np.zeros ((ncol,ngrid), dtype=x[0].dtype)
        rho = rho.sum (0)
        R = self.get_ratio (Pi, rho/2)
        zeta = self.get_zeta (R, fn_deriv=1)
        jTx[:2] = tfnal_derivs._gentLDA_jT_op (x, rho, Pi, R, zeta)
        if self.dens_deriv > 0:
            jTx[:3] += tfnal_derivs._tGGA_jT_op (x, rho, Pi, R, zeta)
        if self.dens_deriv > 1:
            jTx[:4] += tfnal_derivs._tmetaGGA_jT_op(x, rho, Pi, R, zeta)

        return jTx

    def d_jT_op (self, x, rho, Pi):
        r''' Evaluate the x.(nabla j) contribution to the second density
        derivatives of the on-top energy in terms of the untranslated
        density and pair density

        Args:
            x : ndarray of shape (2,*,ngrids)
                Usually, a functional derivative of the on-top xc energy
                wrt translated densities
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]

        Returns: ndarray of shape (*,ngrids)
            second derivative of the translation dotted with x
            3 rows for tLDA and 5 rows for tGGA
        '''
        nrow = 3 + 2*int(self.dens_deriv>0)
        f = np.zeros ((nrow, x[0].shape[-1]), dtype=x[0].dtype)

        rho = rho.sum (0)
        R = self.get_ratio (Pi, rho/2)
        zeta = self.get_zeta (R, fn_deriv=2)

        f[:3] = tfnal_derivs._gentLDA_d_jT_op (x, rho, Pi, R, zeta)
        if self.dens_deriv:
            f[:] += tfnal_derivs._tGGA_d_jT_op (x, rho, Pi, R, zeta)

        if self.verbose >= logger.DEBUG:
            idx = zeta[0] == 0
            logger.debug (self, 'MC-PDFT fot zeta check: %d zeta=0 columns',
                np.count_nonzero (idx))
            if np.count_nonzero (idx):
                for ix, frow in enumerate (f):
                    logger.debug (self, 'MC-PDFT fot zeta check: f[%d] norm '
                        'over zeta=0 columns: %e', ix, linalg.norm (frow[idx]))

        return f

    def eval_ot (self, rho, Pi, dderiv=1, weights=None, _unpack_vot=True):

        eot, vot, fot = tfnal_derivs.eval_ot (self, rho, Pi, dderiv=dderiv,
            weights=weights, _unpack_vot=_unpack_vot)
        if (self.verbose <= logger.DEBUG) or (dderiv<1) or (weights is None):
            return eot, vot, fot
        if rho.ndim == 2: rho = rho[:,None,:]
        if Pi.ndim == 1: Pi = Pi[None,:]
        rho_tot = rho.sum (0)
        nvr = rho_tot.shape[0]
        ngrids = rho_tot.shape[-1]

        r0 = 2*(np.random.rand (ngrids)-1)

        for p in range (20):
            # ~~~ eval_xc reference ~~~
            rho_t0 = self.get_rho_translated (Pi, rho)
            exc, vxc_p, fxc = self._numint.eval_xc (self.otxc, (rho_t0[0,:,:],
                rho_t0[1,:,:]), spin=1, relativity=0, deriv=dderiv,
                verbose=self.verbose)[:3]
            exc *= rho_t0[:,0,:].sum (0)
            vxc_p = tfnal_derivs._reshape_vxc_sigma (vxc_p, self.dens_deriv)
            vxc = tfnal_derivs._unpack_sigma_vector (vxc_p, rho_t0[0,1:4], rho_t0[1,1:4])
            if dderiv>1: fxc = tfnal_derivs._pack_fxc_ltri (fxc, self.dens_deriv)
            # ~~~ shift translated rho directly ~~~
            r = rho_t0 * r0 / 2**p
            drho_t = np.zeros_like (rho_t0, dtype=rho_t0.dtype)
            ndf = 2 * (1 + int (nvr>1))
            drho_t[0,0,0::ndf] = r[0,0,0::ndf]
            drho_t[1,0,1::ndf] = r[1,0,1::ndf]
            if ndf > 2:
                drho_t[0,1:4,2::ndf] = r[0,1:4,2::ndf]
                drho_t[1,1:4,3::ndf] = r[1,1:4,3::ndf]
            # ~~~ eval_xc @ rho_t1 = rho_t0 + drho_t ~~~
            rho_t1 = rho_t0 + drho_t
            exc1, vxc1 = self._numint.eval_xc (self.otxc, (rho_t1[0,:,:],
                rho_t1[1,:,:]), spin=1, relativity=0, deriv=dderiv,
                verbose=self.verbose)[:2]
            exc1 *= rho_t1[:,0,:].sum (0)
            vxc1 =  tfnal_derivs._unpack_vxc_sigma (vxc1, rho_t1, self.dens_deriv)
            df_lbl = ('rhoa', 'rhob', "rhoa'", "rhob'")[:2*(1+int(nvr>1))]
            _v_err_report (self, 'eval_xc {}'.format (p), df_lbl, rho_t0[0], rho_t0[1], exc, vxc,
                vxc_p, fxc, exc1, vxc1, drho_t, weights)

            # ~~~ eval_ot compare ~~~
            nvP = vot[1].shape[0]
            d1 = rho_tot[1:4] if nvr > 1 else None
            d2 = Pi[1:4] if nvP > 1 else None
            if _unpack_vot:
                vot_u = vot
                vot_p = tfnal_derivs.eval_ot (self, rho, Pi, dderiv=dderiv,
                                              weights=weights, _unpack_vot=False)[1]
            else:
                vot_p = vot
                vot_u = tfnal_derivs._unpack_sigma_vector (vot, d1, d2)
            drho = rho_tot * r0 / 2**p
            dPi = Pi * r0 / 2**p
            r, P = drho.copy (), dPi.copy ()
            drho[:] = dPi[:] = 0.0
            ndf = 2 + int(nvr>1) + int(nvP>1)
            drho[0,0::ndf] = r[0,0::ndf]
            dPi[0,1::ndf]  = P[0,1::ndf]
            if ndf > 2: drho[1:4,2::ndf] = r[1:4,2::ndf]
            if ndf > 3:  dPi[1:4,3::ndf] = P[1:4,3::ndf]
            rho1 = rho+(drho/2) # /2 because rho has one more dimension of size = 2
            # that gets summed later
            Pi1 = Pi + dPi
            # ~~~ ignore numerical instability of unfully-translated fnals ~~~
            if self.otxc[0].lower () == 't':
                z0 = self.get_zeta (self.get_ratio (Pi, rho_tot/2)[0],
                    fn_deriv=0)[0]
                z1 = self.get_zeta (self.get_ratio (Pi1, rho1.sum(0)/2)[0],
                    fn_deriv=0)[0]
                idx = (z0==0) |(z1==0)
                drho[:,idx] = dPi[:,idx] = 0
                rho1[:,:,idx] = rho[:,:,idx]
                Pi1[:,idx] = Pi[:,idx]
            # ~~~ eval_ot @ rho1 = rho + drho ~~~
            eot1, vot1 = tfnal_derivs.eval_ot (self, rho1, Pi1,
                dderiv=dderiv, weights=weights, _unpack_vot=True)[:2]
            #vot1 = tfnal_derivs._unpack_sigma_vector (vot1, d1, d2)
            df_lbl = ('rho', 'Pi', "rho'", "Pi'")[:ndf]
            _v_err_report (self, 'eval_ot {}'.format (p), df_lbl, rho_tot, Pi, eot, vot_u, vot_p, fot,
                eot1, vot1, (drho, dPi), weights)

        return eot, vot, fot

    eval_ot.__doc__ = otfnal.eval_ot.__doc__

# TODO: test continuity of smoothing function and warn at initialization?
class ftransfnal (transfnal):
    __doc__ = transfnal.__doc__ + r'''

    Extra attributes for ``fully-translated'' extension of Carlson
    et al., JCTC 11, 4077 (2015):

        R0 : float
            connecting point to polynomial smoothing function;
            R0 <= 1.0. Default is 0.9.
        R1 : float
            endpoint of polynomial smoothing function, zeta(R1) =
            zeta'(R1) = zeta''(R1) = 0.0; R1 >= 1.0. Default is 1.15.
        A : float
            Quintic coefficient of polynomial smoothing function.
            Default = -475.60656009 is chosen to make zeta continuous
            through its second derivative at given the default R0 and R1.
        B : float
            Quartic coefficient of polynomial smoothing function.
            Default = -379.47331922 is chosen to make zeta continuous
            through its second derivative given the default R0 and R1.
        C : float
            Cubic coefficient of polynomial smoothing function.
            Default = -85.38149682 chosen to make zeta continuous
            through its second derivative given the default R0 and R1.
    '''

    transl_prefix='ft'

    def __init__ (self, ks, **kwargs):
        otfnal.__init__(self, ks.mol, **kwargs)
        self.R0=FT_R0
        self.R1=FT_R1
        self.A=FT_A
        self.B=FT_B
        self.C=FT_C
        self.otxc = 'ft' + ks.xc
        self._numint = copy.copy (ks._numint)
        self._numint.libxc = libxc
        self.grids = copy.copy (ks.grids)
        self._numint.hybrid_coeff = ft_hybrid_coeff.__get__(self._numint)
        self._numint.nlc_coeff = ft_nlc_coeff.__get__(self._numint)
        self._numint.rsh_coeff = ft_rsh_coeff.__get__(self._numint)
        self._numint.eval_xc = ft_eval_xc.__get__(self._numint)
        self._numint._xc_type = ft_xc_type.__get__(self._numint)
        self._init_info ()

    Pi_deriv = transfnal.dens_deriv

    def get_rho_translated (self, Pi, rho):
        r''' Compute the "fully-translated" alpha and beta densities
        and their derivatives. This is the same as "translated" except

        rho'_t^a += zeta' * rho / 2
        rho'_t^b -= zeta' * rho / 2

        And the functional form of "zeta" is changed (see "get_zeta")

        Args:
            Pi : ndarray of shape (*, ngrids)
                containing on-top pair density [and derivatives]
            rho : ndarray of shape (2, *, ngrids)
                containing spin density [and derivatives]

        Returns:
            rho_ft : ndarray of shape (2,*,ngrids)
                Fully-translated spin density (and derivatives)
        '''
        nderiv_R = max (rho.shape[1], Pi.shape[0])
        if nderiv_R == 1: return transfnal.get_rho_translated (self, Pi, rho)

        # Spin density and first term of spin gradient in common with transfnal
        rho_avg = (rho[0,:,:] + rho[1,:,:]) / 2
        rho_ft, R, zeta = transfnal.get_rho_translated (self, Pi, rho,
            _fn_deriv=1)

        # Add propagation of chain rule through zeta
        w = (rho_avg[0] * zeta[1])[None,:] * R[1:4]
        rho_ft[0][1:4] += w
        rho_ft[1][1:4] -= w

        return rho_ft

    def get_zeta (self, R, fn_deriv=1):
        r''' Compute the intermediate zeta used to compute the translated spin
        densities and its functional derivatives

        From the "full" translation [Carlson et al., JCTC 11, 4077 (2015)]:
        zeta = (1-R)^(1/2)                          ; R < R0
             = A*(R-R1)^5 + B*(R-R1)^4 + C*(R-R1)^3 ; R0 <= R < R1
             = 0                                    ; otherwise

        Args:
            R : ndarray of shape (*,ngrids)
                Ratio (4Pi/rho^2) and possibly its spatial derivatives
                Only the first row is used in this function

        Kwargs:
            fn_deriv : integer
                order of functional derivative (d^n z / dR^n) to return
                along with the value of zeta

        Returns:
            zeta : ndarray of shape (fn_deriv+1, ngrids)
        '''
        # Rmax unused here. It only needs to be passed in the transfnal version
        if R.ndim == 2: R = R[0]
        R0, R1, A, B, C = self.R0, self.R1, self.A, self.B, self.C
        zeta = transfnal.get_zeta (self, R, fn_deriv=fn_deriv, _Rmax=R0)
        idx = (R >= R0) & (R < R1)
        if not np.count_nonzero (idx): return zeta
        zeta[:,idx] = 0.0
        dR = np.stack ([np.power (R[idx] - R1, n)
            for n in range (1,6)], axis=0)
        def _derivs ():
            yield A*dR[4] +    B*dR[3] +   C*dR[2]
            yield 5*A*dR[3] +  4*B*dR[2] + 3*C*dR[1]
            yield 20*A*dR[2] + 12*B*dR[1] + 6*C*dR[0]
            yield 60*A*dR[1] + 24*B*dR[0] + 6*C
            yield 120*A*dR[0] + 24*B
            yield 120*A
        for n, row in enumerate (_derivs ()):
            zeta[n,idx] = row
            if n == fn_deriv: break

        return zeta

    def jT_op (self, x, rho, Pi, **kwargs):
        r''' Evaluate jTx = (x.j)T where j is the Jacobian of the
        translated densities in terms of the untranslated density and
        pair density

        Args:
            x : ndarray of shape (2,*,ngrids)
                Usually, a functional derivative of the on-top xc energy
                wrt translated densities
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]

        Returns: ndarray of shape (*,ngrids)
            Usually, a functional derivative of the on-top pair density
            exchange-correlation energy wrt to total density and its
            derivatives. The potential must be spin-symmetric in
            pair-density functional theory.
        '''
        ntc = 2 + int(self.dens_deriv>0)
        ncol = 2 + 3*int(self.dens_deriv>0)
        ngrid = rho.shape[-1]
        jTx = np.zeros ((ncol,ngrid), dtype=x[0].dtype)
        jTx[:ntc,:] = transfnal.jT_op (self, x, rho, Pi, **kwargs)
        rho = rho.sum (0)
        R = self.get_ratio (Pi[0:4,:], rho[0:4,:]/2)
        zeta = self.get_zeta (R[0], fn_deriv=2)
        if self.dens_deriv > 0:
            jTx[:] += tfnal_derivs._ftGGA_jT_op (x, rho, Pi, R, zeta)
        return jTx

    def d_jT_op (self, x, rho, Pi, **kwargs):
        r''' Evaluate the x.(nabla j) contribution to the second density
        derivatives of the on-top energy in terms of the untranslated
        density and pair density

        Args:
            x : ndarray of shape (2,*,ngrids)
                Usually, a functional derivative of the on-top xc energy
                wrt translated densities
            rho : ndarray of shape (2,*,ngrids)
                containing spin-density [and derivatives]
            Pi : ndarray with shape (*,ngrids)
                containing on-top pair density [and derivatives]

        Returns: ndarray of shape (*,ngrids)
            second derivative of the translation dotted with x
            3 rows for tLDA and 5 rows for tGGA
        '''
        nrow_t = 3 + 2*int(self.dens_deriv>0)
        nrow = 3 + 12*int(self.dens_deriv>0)
        f = np.zeros ((nrow, x[0].shape[-1]), dtype=x[0].dtype)
        f[:nrow_t] = transfnal.d_jT_op (self, x, rho, Pi, **kwargs)
        if self.dens_deriv:
            rho = rho.sum (0)
            R = self.get_ratio (Pi[0:4,:], rho[0:4,:]/2)
            zeta = self.get_zeta (R[0], fn_deriv=3)
            f[:] += tfnal_derivs._ftGGA_d_jT_op (x, rho, Pi, R, zeta)
        return f


_CS_a_DEFAULT = 0.04918
_CS_b_DEFAULT = 0.132
_CS_c_DEFAULT = 0.2533
_CS_d_DEFAULT = 0.349

def _sanity_check_ftot(xc_code):
    '''
    This function will check the functional type and will
    raise the warning for fully-translated MGGAs or custom functionals.
    '''
    xc_type = libxc.xc_type(xc_code)
    if xc_type not in ['LDA', 'GGA']:
        msg = f"fully-translated {xc_type} on-top functionals are not defined"
        raise NotImplementedError(msg)

def get_transfnal (mol, otxc):
    if otxc.upper () in OT_ALIAS:
        otxc = OT_ALIAS[otxc.upper ()]
    if otxc.upper ().startswith ('T'):
        xc_base = otxc[1:]
        fnal_class = transfnal
    elif otxc.upper ().startswith ('FT'):
        xc_base = otxc[2:]
        _sanity_check_ftot(xc_base)
        fnal_class = ftransfnal
    else:
        raise NotImplementedError (
            'On-top pair-density functional names other than "translated" (t) or '
            '"fully-translated (ft).'
        )
    # Try to register the functional with libxc, if not already done
    xc_base = _get_registered_ot_functional (xc_base, mol)

    xc_base = OT_HYB_ALIAS.get (xc_base.upper (), xc_base)

    if ',' not in xc_base and \
        (xc_base.upper() not in REG_OT_FUNCTIONALS) and \
        (_libxc.is_hybrid_or_rsh (xc_base)):
        raise NotImplementedError (
            'Aliased or built-in translated hybrid or range-separated '
            'functionals\nother than those listed in otfnal.OT_HYB_ALIAS. '
            'Build a compound functional\nstring with a comma separating the '
            'exchange and correlation parts, or use\notfnal.make_hybrid_fnal '
            'instead.'
        )
    ks = dft.RKS (mol)
    ks.xc = xc_base
    return fnal_class (ks)

class colle_salvetti_corr (otfnal):


    def __init__(self, mol, **kwargs):
        super().__init__(mol, **kwargs)
        self.otxc = 'Colle_Salvetti'
        self._numint = NumInt ()
        self.grids = Grids (mol)
        self._numint.hybrid_coeff = lambda * args : 0
        self._numint.nlc_coeff = lambda * args : [0, 0]
        self._numint.rsh_coeff = lambda * args : [0, 0, 0]
        self._numint._xc_type = lambda * args : 'MGGA'
        self.CS_a =_CS_a_DEFAULT
        self.CS_b =_CS_b_DEFAULT
        self.CS_c =_CS_c_DEFAULT
        self.CS_d =_CS_d_DEFAULT
        self._init_info ()

    def get_E_ot (self, rho, Pi, weights):
        r''' Colle & Salvetti, Theor. Chim. Acta 37, 329 (1975)
        see also Lee, Yang, Parr, Phys. Rev. B 37, 785 (1988)
        [Eq. (3)]'''

        a, b, c, d = self.CS_a, self.CS_b, self.CS_c, self.CS_d
        rho_tot = rho[0,0] + rho[1,0]
        idx = rho_tot > 1e-15

        num  = -c * np.power (rho_tot[idx], -1/3)
        num  = np.exp (num, num)
        num *= Pi[4,idx]
        num *= b * np.power (rho_tot[idx], -8/3)
        num += 1

        denom  = d * np.power (rho_tot[idx], -1/3)
        denom += 1

        num /= denom
        num *= Pi[0,idx]
        num /= rho_tot[idx]
        num *= weights[idx]

        E_ot  = np.sum (num)
        E_ot *= -4 * a
        return E_ot

def _hybrid_2c_coeff (ni, xc_code, spin=0):
    ''' Wrapper to the xc_code hybrid coefficient parser to return the
    exchange and correlation components of the hybrid coefficent
    separately '''

    if xc_code.upper() in REG_OT_FUNCTIONALS:
        hyb_x = REG_OT_FUNCTIONALS[xc_code.upper ()].get('hyb_x', 0)
        hyb_c = REG_OT_FUNCTIONALS[xc_code.upper ()].get('hyb_c', 0)
        return [hyb_x, hyb_c]
    else:
        hyb_tot = _NumInt.hybrid_coeff (ni, xc_code, spin=spin)
        if hyb_tot == 0: return [0, 0]

        # For exchange-only functionals, hyb_c = hyb_x
        x_code, c_code = _libxc.split_x_c_comma (xc_code)
        x_code = x_code + ','
        c_code = ',' + c_code

        # All factors of 'HF' are summed by default. Therefore just run the same
        # code for the exchange and correlation parts of the string separately
        hyb_x = _NumInt.hybrid_coeff(ni, x_code, spin=spin) if len (x_code) else 0
        hyb_c = _NumInt.hybrid_coeff(ni, c_code, spin=spin) if len (c_code) else 0
        return [hyb_x, hyb_c]

def make_scaled_fnal (xc_code, hyb_x = 0, hyb_c = 0, fnal_x = None,
        fnal_c = None):
    ''' Convenience function to write the xc_code corresponding to a
        functional of the type

        Exc = hyb_x*E_x[Psi] + fnal_x*E_x[rho] + hyb_c*E_c[Psi]
              + fnal_c*E_c[rho]

        where E[Psi] is an energy from a wave function, and E[rho] is a
        density functional from libxc. The decomposition of E[Psi] into
        exchange (E_x) and correlation (E_c) components is arbitrary.

        Args:
            xc_code : string
                As used in pyscf.dft.libxc. An exception is raised if it
                is already a hybrid or contains a kinetic-energy
                functional component.

        Kwargs:
            hyb_x : float
                fraction of wave function exchange to be included
            hyb_c : float
                fraction of wave function correlation to be included
            fnal_x : float
                fraction of density functional exchange to be included.
                Defaults to 1 - hyb_x.
            fnal_c : float
                fraction of density functional correlation to be
                included. Defaults to 1 - hyb_c.

        returns:
            xc_code : string
                If xc_code has exchange part x_code and correlation part
                c_code, the return value is
                'fnal_x * x_code + hyb_x * HF,
                fnal_c * c_code + hyb_c * HF'
                You STILL HAVE TO PREPEND 't' OR 'ft'!!!
    '''
    if fnal_x is None: fnal_x = 1 - hyb_x
    if fnal_c is None: fnal_c = 1 - hyb_c

    if _libxc.is_hybrid_xc (xc_code):
        raise RuntimeError ('Functional {} is already a hybrid!'.format (
            xc_code))
    x_code, c_code = _libxc.split_x_c_comma (xc_code)

    x_facs, x_terms = _libxc.parse_xc_formula (x_code)
    if fnal_x != 1: x_facs = list (np.asarray (x_facs)*fnal_x)
    if hyb_x != 0:
        x_facs.append (hyb_x)
        x_terms.append ('HF')
    x_code = _libxc.assemble_xc_formula (x_facs, x_terms)

    c_facs, c_terms = _libxc.parse_xc_formula (c_code)
    if fnal_c != 1: c_facs = list (np.asarray (c_facs)*fnal_c)
    if hyb_c != 0:
        c_facs.append (hyb_c)
        c_terms.append ('HF')
    c_code = _libxc.assemble_xc_formula (c_facs, c_terms)

    return x_code + ',' + c_code

def make_hybrid_fnal (xc_code, hyb, hyb_type = 1):
    ''' Convenience function to write "hybrid" xc functional in terms of
        only one parameter

        Args:
            xc_code : string
                As used in pyscf.dft.libxc. An exception is raised if it
                is already a hybrid or contains a kinetic-energy
                functional component.
            hyb : float
                Parameter(s) defining the "hybridization" which is
                handled in various ways according to hyb_type

        Kwargs:
            hyb_type : int or string
                The type of hybrid functional. Current options are:
                - 0 or 'translation': Hybrid fnal is
                    'hyb*HF + (1-hyb)*x_code, hyb*HF + c_code'.
                    Based on the idea that 'exact exchange' of the
                    translated functional corresponds to exchange plus
                    correlation energy of the underlying wave function.
                    Requires len (hyb) == 1.
                - 1 or 'average': Hybrid fnal is
                    'hyb*HF + (1-hyb)*x_code, hyb*HF + (1-hyb)*c_code'.
                    Based on the idea that hyb = 1 recovers the wave
                    function energy itself. Requires len (hyb) == 1.
                - 2 or 'diagram': Hybrid fnal is
                    'hyb*HF + (1-hyb)*x_code, c_code'.
                    Based on the idea that the exchange energy of the
                    wave function somehow can be meaningfully separated
                    from the correlation energy. Requires len (hyb) == 1.
                - 3 or 'lambda': as in arXiv:1911.11162v1. Based on
                    existing 'double-hybrid' functionals. Requires
                    len (hyb) == 1.
                - 4 or 'scaling': Hybrid fnal is
                    'a*HF + (1-a)*x_code, a*HF + (1-a**b)*c_code'
                    where a = hyb[0] and b = 1 + hyb[1]. Based on the
                    scaling inequalities proven by Levy and Perdew in
                    PRA 32, 2010 (1985):
                    E_c[rho_a] < a*E_c[rho] if a < 1 and
                    E_c[rho_a] > a*E_c[rho] if a > 1;
                    BUT
                    E_c[rho_a] ~/~ a^2 E_c[rho], implying that
                    E_c[rho_a] ~ a^b E_c[rho] with b > 1 unknown.
                    Requires len (hyb) == 2.
    '''

    if not hasattr (hyb, '__len__'): hyb = [hyb]
    HYB_TYPE_CODE = {'translation': 0,
                     'average':     1,
                     'diagram':     2,
                     'lambda':      3,
                     'scaling':     4}
    if isinstance (hyb_type, str): hyb_type = HYB_TYPE_CODE[hyb_type]

    if hyb_type == 0:
        assert (len (hyb) == 1)
        return make_scaled_fnal (xc_code, hyb_x=hyb[0], hyb_c=hyb[0],
            fnal_x=(1-hyb[0]), fnal_c=1)
    elif hyb_type == 1:
        assert (len (hyb) == 1)
        return make_scaled_fnal (xc_code, hyb_x=hyb[0], hyb_c=hyb[0],
            fnal_x=(1-hyb[0]), fnal_c=(1-hyb[0]))
    elif hyb_type == 2:
        assert (len (hyb) == 1)
        return make_scaled_fnal (xc_code, hyb_x=hyb[0], hyb_c=0,
            fnal_x=(1-hyb[0]), fnal_c=1)
    elif hyb_type == 3:
        assert (len (hyb) == 1)
        return make_scaled_fnal (xc_code, hyb_x=hyb[0], hyb_c=hyb[0],
            fnal_x=(1-hyb[0]), fnal_c=(1-(hyb[0]*hyb[0])))
    elif hyb_type == 4:
        assert (len (hyb) == 2)
        a = hyb[0]
        b = hyb[0]**(1+hyb[1])
        return make_scaled_fnal (xc_code, hyb_x=a, hyb_c=a, fnal_x=(1-a),
            fnal_c=(1-b))
    else:
        raise RuntimeError ('hybrid type undefined')


# TODO: reconsider this goofy API...
__t_doc__="For 'translated' functionals, otxc string = 't'+xc string\n"
__ft_doc__="For 'fully translated' functionals, otxc string = 'ft'+xc string\n"

def t_hybrid_coeff(ni, xc_code, spin=0):
    #return _NumInt.hybrid_coeff(ni, xc_code[1:], spin=0)
    return _hybrid_2c_coeff (ni, xc_code[1:], spin=0)
t_hybrid_coeff.__doc__ = __t_doc__ + str(_NumInt.hybrid_coeff.__doc__)

def t_nlc_coeff(ni, xc_code):
    return _NumInt.nlc_coeff(ni, xc_code[1:])
t_nlc_coeff.__doc__ = __t_doc__ + str(_NumInt.nlc_coeff.__doc__)

def t_rsh_coeff(ni, xc_code):
    return _NumInt.rsh_coeff(ni, xc_code[1:])
t_rsh_coeff.__doc__ = __t_doc__ + str(_NumInt.rsh_coeff.__doc__)

def t_eval_xc(ni, xc_code, rho, spin=0, relativity=0, deriv=1, verbose=None):
    return _NumInt.eval_xc(ni, xc_code[1:], rho, spin=spin,
        relativity=relativity, deriv=deriv, verbose=verbose)
t_eval_xc.__doc__ = __t_doc__ + str(_NumInt.eval_xc.__doc__)

def t_xc_type(ni, xc_code):
    return _NumInt._xc_type(ni, xc_code[1:])
t_xc_type.__doc__ = __t_doc__ + str(_NumInt._xc_type.__doc__)

def t_rsh_and_hybrid_coeff(ni, xc_code, spin=0):
    return _NumInt.rsh_and_hybrid_coeff (ni, xc_code[1:], spin=spin)
t_rsh_and_hybrid_coeff.__doc__ = (__t_doc__
    + str(_NumInt.rsh_and_hybrid_coeff.__doc__))

def ft_hybrid_coeff(ni, xc_code, spin=0):
    #return _NumInt.hybrid_coeff(ni, xc_code[2:], spin=0)
    return _hybrid_2c_coeff(ni, xc_code[2:], spin=0)
ft_hybrid_coeff.__doc__ = __ft_doc__ + str(_NumInt.hybrid_coeff.__doc__)

def ft_nlc_coeff(ni, xc_code):
    return _NumInt.nlc_coeff(ni, xc_code[2:])
ft_nlc_coeff.__doc__ = __ft_doc__ + str(_NumInt.nlc_coeff.__doc__)

def ft_rsh_coeff(ni, xc_code):
    return _NumInt.rsh_coeff(ni, xc_code[2:])
ft_rsh_coeff.__doc__ = __ft_doc__ + str(_NumInt.rsh_coeff.__doc__)

def ft_eval_xc(ni, xc_code, rho, spin=0, relativity=0, deriv=1, verbose=None):
    return _NumInt.eval_xc(ni, xc_code[2:], rho, spin=spin,
        relativity=relativity, deriv=deriv, verbose=verbose)
ft_eval_xc.__doc__ = __ft_doc__ + str(_NumInt.eval_xc.__doc__)

def ft_xc_type(ni, xc_code):
    return _NumInt._xc_type(ni, xc_code[2:])
ft_xc_type.__doc__ = __ft_doc__ + str(_NumInt._xc_type.__doc__)

def ft_rsh_and_hybrid_coeff(ni, xc_code, spin=0):
    return _NumInt.rsh_and_hybrid_coeff (ni, xc_code[2:], spin=spin)
ft_rsh_and_hybrid_coeff.__doc__ = (__ft_doc__
    + str(_NumInt.rsh_and_hybrid_coeff.__doc__))

def _v_err_report (otfnal, tag, lbls, rho_tot, Pi, e0, v0, v0_packed, f, e1, v1, x, w):
    # Examine the error of the first and second functional derivatives in the
    # debugging block under transfnal.eval_ot below
    logger.debug (otfnal, '--- v_err_report (%s) ---', tag)
    ndf = len (lbls)
    nvP = v0[1].shape[0]
    de = (e1-e0) * w
    vx = ((v0[0]*x[0]).sum (0) + (v0[1]*x[1][:nvP]).sum (0)) * w
    if f is None:
        xfx = np.zeros_like (de)
    else:
        xf = tfnal_derivs.contract_fot (otfnal, f, rho_tot, Pi, x[0], x[1], vot_packed=v0_packed)
        for row in xf: row[:] *= w
        xfx = ((xf[0]*x[0]).sum (0) + (xf[1]*x[1][:nvP]).sum (0)) / 2
        xf_df = [xf[0][0], xf[1][0]]
        dv_df = [(v1[0][0]-v0[0][0])*w, (v1[1][0]-v0[1][0])*w]
        # The lesson of the debug experience from the commented-out block below is:
        # the largest errors (fractional or absolute) in the ftLDA fnal gradient
        # appear to be for R just under 1.0!
        #if 'LDA' in otfnal.otxc:
        #    print ("bigtab", otfnal.otxc, (np.sum (vx) - np.sum (de))/np.sum (de))
        #    tab = np.empty ((xf[0][0].size, 6), dtype=xf[0].dtype)
        #    tab[:,0] = otfnal.get_ratio (Pi, rho_tot/2)[0] - 1.0
        #    tab[:,1] = rho_tot
        #    tab[:,2] = Pi
        #    tab[:,3] = vx
        #    tab[:,4] = tab[:,3] - de
        #    tab[:,5] = tab[:,4] / de
        #    tab[(de==0)&(vx==0),3] = 0.0
        #    tab[(de==0)&(vx!=0),3] = 1.0
        #    tab = tab[np.argsort (-np.abs (tab[:,4])),:]
        #    for row in tab:
        #        print ("{:20.12e} {:9.2e} {:9.2e} {:9.2e} {:9.2e} {:9.2e}".format
        #           (*row))
        if ndf > 2:
            xf_df += [xf[0][1:4].T,]
            dv_df += [((v1[0][1:4]-v0[0][1:4])*w).T,]
        if ndf > 3:
            xf_df += [xf[1][1:4].T,]
            dv_df += [((v1[1][1:4]-v0[1][1:4])*w).T,]
    de_err1 = de - vx
    de_err2 = de_err1 - xfx
    for ix, lbl in enumerate (lbls):
        lib.logger.debug (otfnal, "%s gradient debug %s: %e - %e (- %e) -> %e "
            "(%e)", tag, lbl, np.sum  (de[ix::ndf]), np.sum (vx[ix::ndf]),
            np.sum (xfx[ix::ndf]), np.sum (de_err1[ix::ndf]),
            np.sum (de_err2[ix::ndf]))
    if f is not None:
        for lbl_row, xf_row, dv_row in zip (lbls, xf_df, dv_df):
            err_row = dv_row-xf_row
            for ix_col, lbl_col in enumerate (lbls):
                lib.logger.debug (otfnal, ("%s Hessian debug (H.x_%s)_%s: "
                    "%e - %e -> %e"), tag, lbl_col, lbl_row,
                    linalg.norm (dv_row[ix_col::ndf]),
                    linalg.norm (xf_row[ix_col::ndf]),
                    linalg.norm (err_row[ix_col::ndf]))
                # I am not doing rho'.rho'->sigma right for x.f.x/2
                # However I am somehow doing it right for f.x vs. delta v?
        lib.logger.debug (otfnal, "%s dE - v.x - x.f.x: %e - %e - %e = %e",
            tag, de.sum (), vx.sum (), xfx.sum (), de_err2.sum ())
